import sys
sys.path.append()
import pandas as pd
import numpy as np
from decimal import Decimal
import Functions
from sklearn.model_selection import KFold
from _pydecimal import ROUND_HALF_UP
import random

########################################################################################################################
########################################### Chosen Parameters ##########################################################
########################################################################################################################

# Insert path to the Excel file containing the price data: NaphthaCrackPaper.xlsx
path = input("Input the path to the Excel file NaphthaCrackPaper.xlsx: ")

########################################################################################################################
########################################################################################################################
########################################################################################################################

number_of_splits = 5
validation_type = 'kfold_cross_validation'  # Pick 'kfold_cross_validation' or 'rolling_cross_validation'
Max_Reversion_Period = 5  # Number of days considered after the trigger.
holding_period_range = range(1, Max_Reversion_Period+1, 1)

stop_loss = Decimal(-1.5).quantize(Decimal('0.001'), rounding=ROUND_HALF_UP)
take_profit = Decimal(3).quantize(Decimal('0.001'), rounding=ROUND_HALF_UP)
entry_fee = Decimal('0.1')  # Fee for entering a trade.
exit_fee = Decimal('0.1')  # Fee for exiting a trade.
stop_loss_fee = Decimal('0.2')  # Fee for exiting position in case of stop-loss. Slippage

plot_ts = 1  # Plot the time series (= 1) or not (= 0).
DailyValues, DailyValues_dates, Returns, Returns_dates = Functions.CrackDataImport(Max_Reversion_Period, plot_ts, path)
Returns_dates = Returns_dates.reset_index(drop=True)
Returns_float = [float(x) for x in Returns]

# # Initialising Result DataFrame:
calibration_test_df_cols = ['cross_val_id', 'test_id', 'scaling_type', 'smoothing', 'objective_function', 'strategy',
                            'n_trades_yearly_cali', 'p_value', 'net_profit_pertrade_cali', 'net_profit_yearly_total_cali',
                            'share_pos_cali', 'low_bound', 'holding_period', 'n_trades_yearly_vali',
                            'net_profit_pertrade_vali', 'net_profit_yearly_total_vali', 'share_pos_vali',
                            'net_profit_pertrade_vali_stoptake', 'net_profit_yearly_total_vali_stoptake',
                            'share_pos_vali_stoptake']

calibration_test_df = pd.DataFrame(columns=calibration_test_df_cols)
triggers_test = {}
trades_test = {}
trades_index = {}

kf = KFold(n_splits=number_of_splits, shuffle=False)
splits = list(kf.split(Returns))
std_test = np.zeros(len(splits))
n_trades = 6
n_tries = 10000

for i in range(len(splits)):
    train_index, test_index = splits[i]
    Returns_cali = Returns[train_index]
    Returns_test = Returns[test_index]
    Returns_cali_dates = Returns_dates[train_index].reset_index(drop=True)
    Returns_test_dates = Returns_dates[test_index].reset_index(drop=True) # There lies the problem !!!!!
    delta_matrix = np.zeros((n_tries*n_trades, Max_Reversion_Period+1))
    I = 0
    for j in range(n_tries):
        trades_index = range(0, len(Returns_test) - Max_Reversion_Period, 1)
        trades_trigger = random.choices(trades_index, k=n_trades)
        trades_holding = random.choices(holding_period_range, k=n_trades)
        for k in range(n_trades):
            delta_matrix[I, 0:trades_holding[k]+1] = Returns_test[trades_trigger[k]:trades_trigger[k]+trades_holding[k]+1]
            I = I + 1

    # invert sign for mean reversion
    for j in range(len(delta_matrix)):
        if delta_matrix[j, 0] > 0:
            delta_matrix[j, 1:] = -delta_matrix[j, 1:]

    cum_delta_matrix = delta_matrix[:, 1:].copy()

    for j in range(len(cum_delta_matrix)):
        cum_delta_matrix[j, :] = np.cumsum(cum_delta_matrix[j, :])

    # Rounding and making Decimal:
    cum_delta_matrix = np.round(cum_delta_matrix, 3)
    cum_delta_matrix = np.vectorize(lambda x: Decimal(str(x)))(cum_delta_matrix)
    cum_delta_matrix_stop_take = cum_delta_matrix.copy()

    # Computing the stop loss matrix:
    for k in range(np.shape(cum_delta_matrix_stop_take)[0]):
        for j in range(1, np.shape(cum_delta_matrix_stop_take)[1]):
            if cum_delta_matrix_stop_take[k, j] < stop_loss:
                cum_delta_matrix_stop_take[k, j:] = stop_loss - stop_loss_fee
                break
            if cum_delta_matrix_stop_take[k, j] > take_profit:
                cum_delta_matrix_stop_take[k, j:] = take_profit
                break

    # Adding regular transaction costs:
    cum_delta_matrix_stop_take[:, -1] = cum_delta_matrix_stop_take[:, -1] - entry_fee - exit_fee
    cum_delta_matrix[:, -1] = cum_delta_matrix[:, -1] - entry_fee - exit_fee

    n_trades_yearly_cali = Decimal('0')
    net_profit_pertrade_cali = Decimal('0')
    net_profit_yearly_total_cali = Decimal('0')
    n_trades_yearly_cali = Decimal('0')
    p_value = Decimal('0')
    net_profit_pertrade_cali = Decimal('0')
    net_profit_yearly_total_cali = Decimal('0')
    winrate = Decimal('0')
    n_trades_yearly_vali = Decimal('0')
    net_profit_pertrade_vali = Decimal('0')
    net_profit_yearly_total_vali = Decimal('0')
    share_pos_test = Decimal('0')
    net_profit_pertrade_vali_stoptake = Decimal('0')
    net_profit_yearly_total_vali_stoptake = Decimal('0')
    share_pos_test_stoptake = Decimal('0')
    threshold = Decimal('0')
    holding_period = Decimal('0')

    n_trades_yearly_vali = round(n_trades/(len(Returns_test)/250))
    net_profit_pertrade_vali = np.average(cum_delta_matrix[:, -1])
    net_profit_yearly_total_vali = np.sum(cum_delta_matrix[:, -1])/Decimal(str(round(n_tries/(len(Returns_test)/250), 3)))
    share_pos_test = Decimal(str(round(len(cum_delta_matrix[cum_delta_matrix[:, -1] >= 0, -1])/len(cum_delta_matrix), 3)))
    net_profit_pertrade_vali_stoptake = np.average(cum_delta_matrix_stop_take[:, -1])
    net_profit_yearly_total_vali_stoptake = np.sum(cum_delta_matrix_stop_take[:, -1])/Decimal(str(round(n_tries/(len(Returns_test)/250), 3)))
    share_pos_test_stoptake = Decimal(str(round(len(cum_delta_matrix_stop_take[cum_delta_matrix_stop_take[:, -1] >= 0, -1])/len(cum_delta_matrix_stop_take), 3)))

    calibration_test_row = [['benchmark', i, 'benchmark', 'benchmark', 'benchmark', 'benchmark',
                             n_trades_yearly_cali, p_value, net_profit_pertrade_cali, net_profit_yearly_total_cali,
                             winrate, threshold, holding_period, n_trades_yearly_vali, net_profit_pertrade_vali, net_profit_yearly_total_vali,
                             share_pos_test, net_profit_pertrade_vali_stoptake,
                             net_profit_yearly_total_vali_stoptake, share_pos_test_stoptake]]

    calibration_test_row = pd.DataFrame(calibration_test_row, columns=calibration_test_df_cols)
    calibration_test_df = pd.concat([calibration_test_df, calibration_test_row], ignore_index=True)
