import sys
sys.path.append()
from _pydecimal import ROUND_HALF_UP
from sklearn.model_selection import TimeSeriesSplit
from sklearn.model_selection import KFold
import pandas as pd
import numpy as np
from decimal import Decimal
import Functions
import matplotlib.pyplot as plt
import xlwings as xw


def excel_read_append(file_path, sheet_name, new_rows):

    existing_df = pd.read_excel(file_path, sheet_name=sheet_name)

    if len(existing_df) != 0:
        existing_df = existing_df.iloc[:, 1:]
        new_df = pd.concat([existing_df, new_rows], ignore_index=True)
    else:
        new_df = new_rows

    app = xw.App(visible=True)
    wb = xw.Book(file_path)
    sheet = wb.sheets[sheet_name]
    sheet.clear()
    sheet.range('A1').value = new_df

    wb.save()
    wb.close()
    app.quit()

def scaling_objective_fun(Returns, Returns_dates, window_std, alpha_EWMA, series_choice):

    Std_Dates_Centered, Std_Dates_Backwards, Std_Centered, Std_Centered_EWMA_centered, Std_Backwards, Std_Backwards_EWMA = Functions.RollingSmoothedStdV2(Returns, Returns_dates, window_std, alpha_EWMA, 0)

    output_series = np.zeros(len(Returns))

    # Nap_Std_Centered_full_temp = Nap_Std_Centered_EWMA_cent_full.copy()
    if series_choice == 'std_centered_EWMA_centered':
        chosen_series = Std_Centered_EWMA_centered
        output_series[int(window_std / 2):-int(window_std / 2)] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
        output_series[-int(window_std / 2):] = chosen_series[-1]
    elif series_choice == 'std_backwards_EWMA':
        chosen_series = Std_Backwards_EWMA
        output_series[int(window_std / 2):] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
    elif series_choice == 'std_backwards':
        chosen_series = Std_Backwards
        output_series[int(window_std / 2):] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
    return output_series

def scaling_cali_fun(to_be_scaled_series, scaling_series):
    scaled_series = to_be_scaled_series / scaling_series
    std_cali = np.std(to_be_scaled_series)
    std_cali_scaled = np.std(scaled_series)
    scaled_series = scaled_series / std_cali_scaled * std_cali
    scaled_series = np.array([x.quantize(Decimal('0.001'), rounding=ROUND_HALF_UP) for x in scaled_series])
    return scaled_series, std_cali, std_cali_scaled

def scaling_test_fun(to_be_scaled_series, scaling_series, std_cali, std_cali_scaled):
    scaled_series = to_be_scaled_series / scaling_series
    scaled_series = scaled_series / std_cali_scaled * std_cali
    scaled_series = np.array([x.quantize(Decimal('0.001'), rounding=ROUND_HALF_UP) for x in scaled_series])
    return scaled_series

def FDR_control(p_values_matrix, alpha_FDR, plot_activation):

    vector_p_value = p_values_matrix.flatten()
    sorted_indices = np.argsort(vector_p_value)
    vector_p_value = vector_p_value[sorted_indices]

    Benjamini_val = np.empty(len(vector_p_value))
    n_tests = len(vector_p_value)

    for i in range(len(vector_p_value)):
        Benjamini_val[i] = alpha_FDR*(i+1)/n_tests

    if np.all(np.array([vector_p_value <= Benjamini_val]) == False):
        alpha_FDR_new = 0
    else:
        alpha_FDR_new = np.max(vector_p_value[vector_p_value <= Benjamini_val])
        index = np.argwhere(vector_p_value == alpha_FDR_new)[0][0]

        if plot_activation == 1:
            plt.figure()
            plt.plot(vector_p_value, label='Sorted vector of p-values')
            plt.scatter(range(len(vector_p_value)), vector_p_value)
            plt.plot(Benjamini_val, label='Benjamini-Hochberg criterion')
            plt.scatter(index, alpha_FDR_new, color='red', label='Largest criterion')
            plt.legend()
            plt.xlabel('index of p_values')
            plt.ylabel('value')

    return alpha_FDR_new, vector_p_value

def build_delta_matrix(daily_delta, daily_delta_scaled, max_holding_period, scaling):

    rows = len(daily_delta) - max_holding_period
    cols = max_holding_period + 1
    delta_matrix = np.zeros((rows, cols), dtype=object)

    for k in range(rows):
        if scaling == 1:
            delta_matrix[k, 0] = daily_delta_scaled[k]
        else:
            delta_matrix[k, 0] = daily_delta[k]

        for l in range(2, max_holding_period + 2):
            delta_matrix[k, l - 1] = np.sum(daily_delta[k + 1:k + l])

    # Making the first move always positive, the following cumulative sums positive if reverting, negative if confirming.
    delta_matrix_raw = delta_matrix.copy()
    delta_matrix_raw[:, 0] = daily_delta[:-max_holding_period]
    negative_indices = delta_matrix[:, 0] < 0
    positive_indices = delta_matrix[:, 0] >= 0
    delta_matrix[negative_indices, 0] *= -1
    delta_matrix[positive_indices, 1:] *= -1

    return delta_matrix, delta_matrix_raw

def Calibration(daily_delta, max_holding_period, scaling, daily_delta_scaled, threshold_range, entry_fee, exit_fee, alpha_stat, alpha_FDR, objective_function, strategy, max_signals, plot_activation):

    # Build the matrix of deltas:
    delta_matrix, delta_matrix_raw = build_delta_matrix(daily_delta, daily_delta_scaled, max_holding_period, scaling)

    # Initialising:
    shape = (len(threshold_range), max_holding_period)
    results_pvalue = np.ones(shape, dtype=object)
    results_count = np.zeros(shape, dtype=object)
    results_winrate = np.zeros(shape, dtype=object)
    results_raw = np.zeros(shape, dtype=object)
    results_avg = np.zeros(shape, dtype=object)
    results_tot = np.zeros(shape, dtype=object)
    results_tot_winrate = np.zeros(shape, dtype=object)
    results_sharpe = np.zeros(shape, dtype=object)

    # Applying the strategy to the delta_matrix returns
    if strategy == 2:
        delta_matrix_strat = Functions.Strategy_2_Fun(delta_matrix)
    elif strategy == 3:
        delta_matrix_strat = Functions.Strategy_3_Fun(delta_matrix)
    else:
        delta_matrix_strat = Functions.Strategy_1_Fun(delta_matrix)

    for i in range(len(threshold_range)):

        threshold = threshold_range[i]
        delta_matrix_signal = delta_matrix_strat[delta_matrix_strat[:, 0] >= threshold, :]
        signal_count = len(delta_matrix_signal[:, 0])

        if not delta_matrix_signal.size == 0:

            for j in range(max_holding_period):

                holding_period = j + 1
                p_value = Functions.SignificativeReversion(delta_matrix, delta_matrix_signal, holding_period, 'one-tailed-upper')

                # Computing average return a.k.a. per trade return and subtracting the costs of trading:
                average_return = np.average(delta_matrix_signal[:, holding_period]) - exit_fee - entry_fee
                average_return = average_return.quantize(Decimal('0.001'), rounding=ROUND_HALF_UP)

                # Computing the total returns:
                total_return = average_return * signal_count

                # Computing the winrate:
                winrate = round(len(delta_matrix_signal[delta_matrix_signal[:, holding_period] >= 0, holding_period]) / len(delta_matrix_signal[:, holding_period]), 3)
                winrate = Decimal(str(winrate))
                results_winrate[i, j] = winrate

                # Adding transaction costs for raw returns:
                results_raw[i, j] = average_return + exit_fee + entry_fee
                # Per trade average:
                results_avg[i, j] = average_return
                # Total returns:
                results_tot[i, j] = total_return
                # Include winrate:
                results_tot_winrate[i, j] = total_return * winrate
                # Alternative Metric
                if ((strategy == 2) | (strategy == 3)) & (holding_period == 1):
                    results_sharpe[i, j] = Decimal('0')
                else:
                    if len(delta_matrix_signal[:, holding_period]) > 1:
                        # results_sharpe[i, j] = average_return / Decimal(str(np.std(delta_matrix_signal[:, holding_period])))
                        results_sharpe[i, j] = average_return * Decimal(str(np.sqrt(signal_count))) / Decimal(str(np.std(delta_matrix_signal[:, holding_period])))
                    else:
                        results_sharpe[i, j] = average_return
                # Fill in matrix for p_values:
                results_pvalue[i, j] = p_value
                # Fill in vector for count:
                results_count[i, j] = signal_count

    if (strategy == 2) | (strategy == 3):
        results_pvalue_fdr = results_pvalue[:, 1:].copy()
    else:
        results_pvalue_fdr = results_pvalue.copy()

    # Compute threshold for FDR control:
    alpha_FDR_new, vector_p_value = FDR_control(results_pvalue_fdr, alpha_FDR, plot_activation)

    # Applying the thresholds:
    results_raw_filtered = results_raw.copy()
    results_raw_filtered[(results_pvalue > alpha_FDR_new) | (results_count > max_signals) | (results_pvalue > alpha_stat)] = 0
    results_avg_filtered = results_avg.copy()
    results_avg_filtered[(results_pvalue > alpha_FDR_new) | (results_count > max_signals) | (results_pvalue > alpha_stat)] = 0
    results_tot_filtered = results_tot.copy()
    results_tot_filtered[(results_pvalue > alpha_FDR_new) | (results_count > max_signals) | (results_pvalue > alpha_stat)] = 0
    results_tot_winrate_filtered = results_tot_winrate.copy()
    results_tot_winrate_filtered[(results_pvalue > alpha_FDR_new) | (results_count > max_signals) | (results_pvalue > alpha_stat)] = 0
    results_sharpe_filtered = results_sharpe.copy()
    results_sharpe_filtered[(results_pvalue > alpha_FDR_new) | (results_count > max_signals) | (results_pvalue > alpha_stat)] = 0

    if np.sum(results_raw_filtered) != 0:
        if objective_function == 1:
            max_indices = np.argwhere(results_avg_filtered == np.max(results_avg_filtered))
        elif objective_function == 2:
            max_indices = np.argwhere(results_tot_winrate_filtered == np.max(results_tot_winrate_filtered))
        elif objective_function == 3:
            max_indices = np.argwhere(results_sharpe_filtered == np.max(results_sharpe_filtered))
        else:
            max_indices = np.argwhere(results_tot_filtered == np.max(results_tot_filtered))

        max_indices = max_indices[0]
        best_0 = max_indices[0]
        best_1 = max_indices[1]
        n_vals = results_count[best_0, best_1]
        p_value = results_pvalue[best_0, best_1]
        return_avg = results_avg[best_0, best_1]
        return_tot = results_tot[best_0, best_1]
        winrate = results_winrate[best_0, best_1]
        threshold = threshold_range[best_0]
        holding_period = best_1 + 1

    else:
        print('No statistically significant result found for this combination of parameters.')
        max_indices = 0
        n_vals = 0
        p_value = 0
        return_avg = 0
        return_tot = 0
        winrate = 0
        threshold = 0
        holding_period = 0

    return max_indices, n_vals, p_value, return_avg, return_tot, winrate, threshold, holding_period

def Test(threshold, holding_period, delta_test, scaling, delta_test_scaled, strategy, stoptake_act, stop_loss, take_profit, entry_fee, exit_fee, stop_loss_fee):

    delta_matrix, delta_matrix_raw = build_delta_matrix(delta_test, delta_test_scaled, holding_period, scaling)

    if strategy == 2:
        delta_matrix_strat = Functions.Strategy_2_Fun(delta_matrix)
    elif strategy == 3:
        delta_matrix_strat = Functions.Strategy_3_Fun(delta_matrix)
    else:
        delta_matrix_strat = Functions.Strategy_1_Fun(delta_matrix)

    indexes = np.arange(len(delta_test) - holding_period)
    delta_matrix_signal = delta_matrix_strat[delta_matrix_strat[:, 0] >= threshold, :]
    unscaled_signal = delta_matrix_raw[delta_matrix_strat[:, 0] >= threshold, 0]

    if stoptake_act == 1:
        for i in range(np.shape(delta_matrix_signal)[0]):
            for j in range(1, np.shape(delta_matrix_signal)[1]):
                if delta_matrix_signal[i, j] < stop_loss:
                    delta_matrix_signal[i, j:] = stop_loss - stop_loss_fee
                    break
                if delta_matrix_signal[i, j] > take_profit:
                    delta_matrix_signal[i, j:] = take_profit
                    break

    scaled_signal = delta_matrix_signal[:, 0]
    signal_returns = delta_matrix_signal[:, holding_period] - entry_fee - exit_fee
    signal_indexes = indexes[np.array(delta_matrix_strat[:, 0] >= threshold)]
    vali_df = pd.DataFrame({'index': signal_indexes, 'unscaled_signal': unscaled_signal, 'scaled_signal': scaled_signal, 'signal_returns': signal_returns})

    return vali_df


########################################################################################################################
########################################################################################################################
########################################### Function Parameters And Inputs #############################################
########################################################################################################################
########################################################################################################################

########################################################################################################################
########################################### Chosen Parameters ##########################################################
########################################################################################################################

# Insert path to the Excel file containing the price data: NaphthaCrackPaper.xlsx
path = input("Input the path to the Excel file NaphthaCrackPaper.xlsx: ")

# Select the standardising series: 'backwards std 20 days', 'backwards std 60 days', 'GARCH(1,1)'
scaling_type = input("Select the standardising series: backwards std 20 days, backwards std 60 days, GARCH(1,1): ")
# Select the smoothing coefficient:
alpha_EWMA = float(input("Select the EWMA smoothing coefficient (base case is 0.2): "))
# Select the performance measure:
objective_function = int(input("Select performance measure: 0 for total profit, 1 for profit per trade: "))
# Select the core strategy:
Strategy = int(input("Select performance measure: 1 for direct entry, 2 for delayed entry: "))


########################################################################################################################
########################################### Fixed Parameters ###########################################################
########################################################################################################################

stat_sig_threshold = Decimal('0.05')  # Threshold for statistical significance.
alpha_FDR = Decimal('0.1')
number_of_splits = 5
validation_type = 'kfold_cross_validation'  # Pick 'kfold_cross_validation' or 'rolling_cross_validation'
cross_val_id = 0  # Cross-validation test id.
ScalingType = 1  # 0: No Scaling, 1: Scaling Models
Max_Reversion_Period = 5  # Number of days considered after the trigger.
Min_CoolDown_Period = 1.5  # Min number of weeks between each trade on average.
take_profit_after = Decimal('3')
stop_loss_after = Decimal('-1.5')
stop_loss = Decimal(-1.5).quantize(Decimal('0.001'), rounding=ROUND_HALF_UP)
take_profit = Decimal(3).quantize(Decimal('0.001'), rounding=ROUND_HALF_UP)
excel_export_path = r'F:/TTF/Briac Thesis/RobustnessResultsV4.xlsx'

########################################################################################################################
########################################### Increments and Ranges Definition ###########################################
########################################################################################################################

step_size = Decimal('0.1')  # Step size between values
entry_fee = Decimal('0.1')  # Fee for entering a trade.
exit_fee = Decimal('0.1')  # Fee for exiting a trade.
stop_loss_fee = Decimal('0.2')  # Fee for exiting position in case of stop-loss. Slippage

# Lower bound:
lower_start_value = Decimal('0.8')  # Start value of the range
lower_stop_value = Decimal('2.1')  # Stop value of the range = lower_stop_value - step_size
lo_bound_range = np.arange(lower_start_value, lower_stop_value, step_size)

########################################################################################################################
########################################### Data Import and Selection ##################################################
########################################################################################################################

if ScalingType == 0:
    Scaling = 0
else:
    Scaling = 1

if objective_function == 0:
    optimisation_objective = 'total returns: triggers*avg(returns)'
elif objective_function == 1:
    optimisation_objective = 'average returns: avg(returns)'

plot_ts = 1  # Plot the time series (= 1) or not (= 0).
DailyValues, DailyValues_dates, Returns, Returns_dates = Functions.CrackDataImport(Max_Reversion_Period, plot_ts, path)

Returns_dates = Returns_dates.reset_index(drop=True)
Returns_float = [float(x) for x in Returns]

# Maximum Opportunities for the given time-series. Set to 1 Trade per Min_CoolDown_Period [weeks].
Max_Opportunities = int((len(Returns)/5)/Min_CoolDown_Period)

smoothing = f'EWMA alpha={alpha_EWMA}'

########################################################################################################################
########################################################################################################################
################################################ Robustness Check ######################################################
########################################################################################################################
########################################################################################################################

# Initialising Result DataFrame:
calibration_test_df_cols = ['cross_val_id', 'test_id', 'scaling_type', 'smoothing', 'objective_function', 'strategy',
                            'n_trades_yearly_cali', 'p_value', 'net_profit_pertrade_cali', 'net_profit_yearly_total_cali',
                            'share_pos_cali', 'low_bound', 'holding_period', 'n_trades_yearly_vali',
                            'net_profit_pertrade_vali', 'net_profit_yearly_total_vali', 'share_pos_vali',
                            'net_profit_pertrade_vali_stoptake', 'net_profit_yearly_total_vali_stoptake',
                            'share_pos_vali_stoptake']

calibration_test_df = pd.DataFrame(columns=calibration_test_df_cols)

triggers_test = {}
trades_test = {}
trades_index = {}

if validation_type == 'rolling_cross_validation':
    tscv = TimeSeriesSplit(n_splits=number_of_splits)
    splits = list(tscv.split(Returns))

elif validation_type == 'kfold_cross_validation':
    kf = KFold(n_splits=number_of_splits, shuffle=False)
    splits = list(kf.split(Returns))

std_test = np.zeros(len(splits))
i = 0
for i in range(len(splits)):

    train_index, test_index = splits[i]
    Returns_cali = Returns[train_index]
    Returns_test = Returns[test_index]
    Returns_cali_dates = Returns_dates[train_index].reset_index(drop=True)
    Returns_test_dates = Returns_dates[test_index].reset_index(drop=True)

    test_end = test_index[-1]
    cali_end = train_index[-1]
    std_test[i] = np.std(Returns_test)

    ########################################################################################################################
    ################################################## Calibration #########################################################
    ########################################################################################################################

    # Scaling Returns_cali
    if ScalingType == 0:
        Returns_cali_scaled = Returns_cali.copy()

    elif ScalingType == 1:

        if scaling_type == 'backwards std 20 days':
            scaling_series_float = scaling_objective_fun(Returns_cali, Returns_cali_dates, 40, alpha_EWMA, 'std_backwards_EWMA')
            scaling_series = np.array([Decimal(x) for x in scaling_series_float])
            Returns_cali_scaled, std_cali, std_cali_scaled = scaling_cali_fun(Returns_cali, scaling_series)

        if scaling_type == 'backwards std 60 days':
            scaling_series_float = scaling_objective_fun(Returns_cali, Returns_cali_dates, 120, alpha_EWMA, 'std_backwards_EWMA')
            scaling_series = np.array([Decimal(x) for x in scaling_series_float])
            Returns_cali_scaled, std_cali, std_cali_scaled = scaling_cali_fun(Returns_cali, scaling_series)

        if scaling_type == 'garch 1,1':
            Returns_cali_float = [float(x) for x in Returns_cali]

            # Garch Parameters:
            pred_horizon = 1
            p_param = 1
            q_param = 1
            garch_stabilisation_delay = 200

            # Computing Garch:
            Garch_Implied_Vol_full = Functions.Rolling_Garch(Returns_cali_float, pred_horizon, p_param, q_param)
            Garch_Implied_Vol = Garch_Implied_Vol_full[garch_stabilisation_delay:]
            Garch_dates = Returns_cali_dates[garch_stabilisation_delay:]
            Returns_cali = Returns_cali[garch_stabilisation_delay:]
            scaling_series_float = Functions.EWMA(Garch_Implied_Vol, alpha_EWMA)
            scaling_series = np.array([Decimal(x) for x in scaling_series_float])
            Returns_cali_scaled, std_cali, std_cali_scaled = scaling_cali_fun(Returns_cali, scaling_series)

    max_indices, n_vals, p_value, return_avg, return_tot, winrate, threshold, holding_period = Calibration(Returns_cali.copy(), Max_Reversion_Period, Scaling, Returns_cali_scaled.copy(), lo_bound_range, entry_fee, exit_fee, stat_sig_threshold, alpha_FDR, objective_function, Strategy, Max_Opportunities, 0)

    if n_vals == 0:
        n_trades_yearly_cali = Decimal('0')
        net_profit_pertrade_cali = Decimal('0')
        net_profit_yearly_total_cali = Decimal('0')
        n_trades_yearly_cali = Decimal('0')
        p_value = Decimal('0')
        net_profit_pertrade_cali = Decimal('0')
        net_profit_yearly_total_cali = Decimal('0')
        winrate = Decimal('0')
        threshold = Decimal('0')
        holding_period = Decimal('0')
        n_trades_yearly_vali = Decimal('0')
        net_profit_pertrade_vali = Decimal('0')
        net_profit_yearly_total_vali = Decimal('0')
        share_pos_test = Decimal('0')
        net_profit_pertrade_vali_stoptake = Decimal('0')
        net_profit_yearly_total_vali_stoptake = Decimal('0')
        share_pos_test_stoptake = Decimal('0')
    else:
        # Preparing results for output:
        n_trades_yearly_cali = Decimal(str(round(n_vals / len(Returns_cali) * 250, 2)))
        net_profit_pertrade_cali = Decimal(str(round(return_avg, 3)))
        net_profit_yearly_total_cali = Decimal(str(round(return_tot / len(Returns_cali) * 250, 2)))


        ########################################################################################################################
        ###################################################### Test  ###########################################################
        ########################################################################################################################

        # Scaling Returns_test
        if ScalingType == 0:
            Returns_test_scaled = Returns_test.copy()

        else:
            if scaling_type == 'backwards std 20 days':
                # Get the full time series:
                Returns_test_temp = np.append(Returns_cali, Returns_test)
                Returns_test_temp_dates = np.append(Returns_cali_dates, Returns_test_dates)

                # create the scaling series:
                scaling_series_float = scaling_objective_fun(Returns_test_temp, Returns_test_temp_dates, 40, alpha_EWMA, 'std_backwards_EWMA')
                scaling_series_float = scaling_series_float[len(Returns_cali):]
                scaling_series = np.array([Decimal(x) for x in scaling_series_float])
                Returns_test_scaled = scaling_test_fun(Returns_test, scaling_series, std_cali, std_cali_scaled)

            elif scaling_type == 'backwards std 60 days':
                # Get the full time series:
                Returns_test_temp = np.append(Returns_cali, Returns_test)
                Returns_test_temp_dates = np.append(Returns_cali_dates, Returns_test_dates)

                # create the scaling series:
                scaling_series_float = scaling_objective_fun(Returns_test_temp, Returns_test_temp_dates, 120, alpha_EWMA, 'std_backwards_EWMA')
                scaling_series_float = scaling_series_float[len(Returns_cali):]
                scaling_series = np.array([Decimal(x) for x in scaling_series_float])
                Returns_test_scaled = scaling_test_fun(Returns_test, scaling_series, std_cali, std_cali_scaled)

            elif scaling_type == 'garch 1,1':
                # Get the full time series:
                Returns_test_temp = np.append(Returns_cali, Returns_test)
                Returns_test_temp_float = [float(x) for x in Returns_test_temp]

                # Computing Garch Vol and smoothing it using wavelets:
                Garch_Implied_Vol = Functions.Rolling_Garch(Returns_test_temp_float, pred_horizon, p_param, q_param)
                Garch_Implied_Vol[0] = Garch_Implied_Vol[1]
                # Further smooth using EWMA:
                Garch_Implied_Vol_smoothed = Functions.EWMA(Garch_Implied_Vol, alpha_EWMA)
                # Keep only test period:
                Garch_Implied_Vol = Garch_Implied_Vol[len(Returns_cali):]
                scaling_series_float = Garch_Implied_Vol_smoothed[len(Returns_cali):]
                scaling_series = np.array([Decimal(x) for x in scaling_series_float])

                # scaling the test data set:
                Returns_test_scaled = scaling_test_fun(Returns_test, scaling_series, std_cali, std_cali_scaled)

        vali_df = Test(threshold, holding_period, Returns_test, Scaling, Returns_test_scaled, Strategy, 0, stop_loss, take_profit, entry_fee, exit_fee, stop_loss_fee)
        vali_df_stoptake = Test(threshold, holding_period, Returns_test, Scaling, Returns_test_scaled, Strategy, 1, stop_loss, take_profit, entry_fee, exit_fee, stop_loss_fee)

        n_vals_test = len(vali_df)
        if n_vals_test > 0:
            trade_returns = np.array(vali_df['signal_returns'])
            trade_returns_stoptake = np.array(vali_df_stoptake['signal_returns'])
            triggers = np.array(vali_df['unscaled_signal'])
            index = np.array(vali_df['index'])

            return_tot_test = np.sum(trade_returns)
            return_tot_test_stoptake = np.sum(trade_returns_stoptake)
            return_avg_test = np.average(trade_returns)
            return_avg_test_stoptake = np.average(trade_returns_stoptake)
            trades_test[i] = trade_returns
            triggers_test[i] = triggers
            trades_index[i] = index
            share_pos_test = Decimal(str(round(len(trade_returns[trade_returns >= 0])/len(trade_returns), 3)))
            share_pos_test_stoptake = Decimal(str(round(len(trade_returns_stoptake[trade_returns_stoptake >= 0])/len(trade_returns_stoptake), 3)))

            n_trades_yearly_vali = Decimal(str(round(n_vals_test / len(Returns_test) * 250, 2)))
            net_profit_pertrade_vali = Decimal(str(round(return_avg_test, 3)))
            net_profit_yearly_total_vali = Decimal(str(round(return_tot_test / len(Returns_test) * 250, 2)))
            net_profit_pertrade_vali_stoptake = Decimal(str(round(return_avg_test_stoptake, 3)))
            net_profit_yearly_total_vali_stoptake = Decimal(str(round(return_tot_test_stoptake / len(Returns_test) * 250, 2)))

        else:
            return_tot_test = Decimal('0')
            return_tot_test_stoptake = Decimal('0')
            return_avg_test = Decimal('0')
            return_avg_test_stoptake = Decimal('0')
            trades_test[i] = np.array(Decimal('0'))
            triggers_test[i] = np.array(Decimal('0'))
            trades_index[i] = np.array(Decimal('0'))
            share_pos_test = Decimal('0')
            share_pos_test_stoptake = Decimal('0')

            n_trades_yearly_vali = Decimal('0')
            net_profit_pertrade_vali = Decimal('0')
            net_profit_yearly_total_vali = Decimal('0')
            net_profit_pertrade_vali_stoptake = Decimal('0')
            net_profit_yearly_total_vali_stoptake = Decimal('0')

    calibration_test_row = [[cross_val_id, i, scaling_type, smoothing, objective_function, Strategy,
                             n_trades_yearly_cali, p_value, net_profit_pertrade_cali, net_profit_yearly_total_cali,
                             winrate, threshold, holding_period, n_trades_yearly_vali, net_profit_pertrade_vali,
                             net_profit_yearly_total_vali, share_pos_test, net_profit_pertrade_vali_stoptake,
                             net_profit_yearly_total_vali_stoptake, share_pos_test_stoptake]]

    calibration_test_row = pd.DataFrame(calibration_test_row, columns=calibration_test_df_cols)
    calibration_test_df = pd.concat([calibration_test_df, calibration_test_row], ignore_index=True)

# Create Summary Row for all tests of the cross-validation
calibration_test_df_nz = calibration_test_df[calibration_test_df['n_trades_yearly_cali'] != 0]

p_value_avg = np.mean(calibration_test_df_nz['p_value'])
p_value_med = np.median(calibration_test_df_nz['p_value'])
p_value_min = np.min(calibration_test_df_nz['p_value'])
p_value_max = np.max(calibration_test_df_nz['p_value'])

n_trades_yearly_cali_avg = np.mean(calibration_test_df_nz['n_trades_yearly_cali'])
n_trades_yearly_cali_med = np.median(calibration_test_df_nz['n_trades_yearly_cali'])
n_trades_yearly_cali_min = np.min(calibration_test_df_nz['n_trades_yearly_cali'])
n_trades_yearly_cali_max = np.max(calibration_test_df_nz['n_trades_yearly_cali'])

net_profit_pertrade_cali_avg = np.mean(calibration_test_df_nz['net_profit_pertrade_cali'])
net_profit_pertrade_cali_med = np.median(calibration_test_df_nz['net_profit_pertrade_cali'])
net_profit_pertrade_cali_min = np.min(calibration_test_df_nz['net_profit_pertrade_cali'])
net_profit_pertrade_cali_max = np.max(calibration_test_df_nz['net_profit_pertrade_cali'])

net_profit_yearly_total_cali_avg = np.mean(calibration_test_df_nz['net_profit_yearly_total_cali'])
net_profit_yearly_total_cali_med = np.median(calibration_test_df_nz['net_profit_yearly_total_cali'])
net_profit_yearly_total_cali_min = np.min(calibration_test_df_nz['net_profit_yearly_total_cali'])
net_profit_yearly_total_cali_max = np.max(calibration_test_df_nz['net_profit_yearly_total_cali'])

share_pos_cali_avg = np.average(calibration_test_df_nz['share_pos_cali'], weights=calibration_test_df_nz['n_trades_yearly_cali'])
share_pos_cali_med = np.median(calibration_test_df_nz['share_pos_cali'])
share_pos_cali_min = np.min(calibration_test_df_nz['share_pos_cali'])
share_pos_cali_max = np.max(calibration_test_df_nz['share_pos_cali'])

low_bound_avg = np.mean(calibration_test_df_nz['low_bound'])
low_bound_med = np.median(calibration_test_df_nz['low_bound'])
low_bound_min = np.min(calibration_test_df_nz['low_bound'])
low_bound_max = np.max(calibration_test_df_nz['low_bound'])

holding_period_avg = np.mean(calibration_test_df_nz['holding_period'])
holding_period_med = np.median(calibration_test_df_nz['holding_period'])
holding_period_min = np.min(calibration_test_df_nz['holding_period'])
holding_period_max = np.max(calibration_test_df_nz['holding_period'])

n_trades_yearly_vali_avg = np.mean(calibration_test_df_nz['n_trades_yearly_vali'])
n_trades_yearly_vali_med = np.median(calibration_test_df_nz['n_trades_yearly_vali'])
n_trades_yearly_vali_min = np.min(calibration_test_df_nz['n_trades_yearly_vali'])
n_trades_yearly_vali_max = np.max(calibration_test_df_nz['n_trades_yearly_vali'])

net_profit_pertrade_vali_avg = np.mean(calibration_test_df_nz['net_profit_pertrade_vali'])
net_profit_pertrade_vali_med = np.median(calibration_test_df_nz['net_profit_pertrade_vali'])
net_profit_pertrade_vali_min = np.min(calibration_test_df_nz['net_profit_pertrade_vali'])
net_profit_pertrade_vali_max = np.max(calibration_test_df_nz['net_profit_pertrade_vali'])

net_profit_yearly_total_vali_avg = np.mean(calibration_test_df_nz['net_profit_yearly_total_vali'])
net_profit_yearly_total_vali_med = np.median(calibration_test_df_nz['net_profit_yearly_total_vali'])
net_profit_yearly_total_vali_min = np.min(calibration_test_df_nz['net_profit_yearly_total_vali'])
net_profit_yearly_total_vali_max = np.max(calibration_test_df_nz['net_profit_yearly_total_vali'])

share_pos_vali_avg = np.average(calibration_test_df_nz['share_pos_vali'], weights=calibration_test_df_nz['n_trades_yearly_vali'])
share_pos_vali_med = np.median(calibration_test_df_nz['share_pos_vali'])
share_pos_vali_min = np.min(calibration_test_df_nz['share_pos_vali'])
share_pos_vali_max = np.max(calibration_test_df_nz['share_pos_vali'])

net_profit_pertrade_vali_stoptake_avg = np.mean(calibration_test_df_nz['net_profit_pertrade_vali_stoptake'])
net_profit_pertrade_vali_stoptake_med = np.median(calibration_test_df_nz['net_profit_pertrade_vali_stoptake'])
net_profit_pertrade_vali_stoptake_min = np.min(calibration_test_df_nz['net_profit_pertrade_vali_stoptake'])
net_profit_pertrade_vali_stoptake_max = np.max(calibration_test_df_nz['net_profit_pertrade_vali_stoptake'])

net_profit_yearly_total_vali_stoptake_avg = np.mean(calibration_test_df_nz['net_profit_yearly_total_vali_stoptake'])
net_profit_yearly_total_vali_stoptake_med = np.median(calibration_test_df_nz['net_profit_yearly_total_vali_stoptake'])
net_profit_yearly_total_vali_stoptake_min = np.min(calibration_test_df_nz['net_profit_yearly_total_vali_stoptake'])
net_profit_yearly_total_vali_stoptake_max = np.max(calibration_test_df_nz['net_profit_yearly_total_vali_stoptake'])

share_pos_vali_stoptake_avg = np.mean(calibration_test_df_nz['share_pos_vali_stoptake'])
share_pos_vali_stoptake_med = np.median(calibration_test_df_nz['share_pos_vali_stoptake'])
share_pos_vali_stoptake_min = np.min(calibration_test_df_nz['share_pos_vali_stoptake'])
share_pos_vali_stoptake_max = np.max(calibration_test_df_nz['share_pos_vali_stoptake'])


summary_df_cols = ['cross_val_id', 'scaling_type', 'smoothing', 'objective_function', 'strategy',
                   'n_trades_yearly_cali_avg', 'n_trades_yearly_cali_med', 'n_trades_yearly_cali_min',
                   'n_trades_yearly_cali_max', 'p_value_avg', 'p_value_med', 'p_value_min', 'p_value_max',
                   'net_profit_pertrade_cali_avg',
                   'net_profit_pertrade_cali_med', 'net_profit_pertrade_cali_min', 'net_profit_pertrade_cali_max',
                   'net_profit_yearly_total_cali_avg', 'net_profit_yearly_total_cali_med',
                   'net_profit_yearly_total_cali_min', 'net_profit_yearly_total_cali_max',
                   'share_pos_cali_avg', 'share_pos_cali_med', 'share_pos_cali_min', 'share_pos_cali_max',
                   'low_bound_avg', 'low_bound_med', 'low_bound_min', 'low_bound_max', 'holding_period_avg',
                   'holding_period_med', 'holding_period_min', 'holding_period_max', 'n_trades_yearly_vali_avg',
                   'n_trades_yearly_vali_med', 'n_trades_yearly_vali_min', 'n_trades_yearly_vali_max',
                   'net_profit_pertrade_vali_avg', 'net_profit_pertrade_vali_med',
                   'net_profit_pertrade_vali_min', 'net_profit_pertrade_vali_max',
                   'net_profit_yearly_total_vali_avg', 'net_profit_yearly_total_vali_med',
                   'net_profit_yearly_total_vali_min', 'net_profit_yearly_total_vali_max',
                   'share_pos_vali_avg', 'share_pos_vali_med', 'share_pos_vali_min', 'share_pos_vali_max',
                   'net_profit_pertrade_vali_stoptake_avg', 'net_profit_pertrade_vali_stoptake_med',
                   'net_profit_pertrade_vali_stoptake_min', 'net_profit_pertrade_vali_stoptake_max',
                   'net_profit_yearly_total_vali_stoptake_avg', 'net_profit_yearly_total_vali_stoptake_med',
                   'net_profit_yearly_total_vali_stoptake_min', 'net_profit_yearly_total_vali_stoptake_max',
                   'share_pos_vali_stoptake_avg', 'share_pos_vali_stoptake_med', 'share_pos_vali_stoptake_min',
                   'share_pos_vali_stoptake_max']


summary_df_row = [[cross_val_id, scaling_type, smoothing, objective_function, Strategy,
                   n_trades_yearly_cali_avg, n_trades_yearly_cali_med, n_trades_yearly_cali_min,
                   n_trades_yearly_cali_max, p_value_avg, p_value_med, p_value_min, p_value_max,
                   net_profit_pertrade_cali_avg, net_profit_pertrade_cali_med, net_profit_pertrade_cali_min,
                   net_profit_pertrade_cali_max, net_profit_yearly_total_cali_avg,
                   net_profit_yearly_total_cali_med, net_profit_yearly_total_cali_min,
                   net_profit_yearly_total_cali_max, share_pos_cali_avg, share_pos_cali_med, share_pos_cali_min,
                   share_pos_cali_max, low_bound_avg, low_bound_med, low_bound_min, low_bound_max,
                   holding_period_avg, holding_period_med, holding_period_min, holding_period_max,
                   n_trades_yearly_vali_avg, n_trades_yearly_vali_med, n_trades_yearly_vali_min,
                   n_trades_yearly_vali_max, net_profit_pertrade_vali_avg, net_profit_pertrade_vali_med,
                   net_profit_pertrade_vali_min, net_profit_pertrade_vali_max, net_profit_yearly_total_vali_avg,
                   net_profit_yearly_total_vali_med, net_profit_yearly_total_vali_min,
                   net_profit_yearly_total_vali_max, share_pos_vali_avg, share_pos_vali_med, share_pos_vali_min,
                   share_pos_vali_max, net_profit_pertrade_vali_stoptake_avg, net_profit_pertrade_vali_stoptake_med,
                   net_profit_pertrade_vali_stoptake_min, net_profit_pertrade_vali_stoptake_max,
                   net_profit_yearly_total_vali_stoptake_avg, net_profit_yearly_total_vali_stoptake_med,
                   net_profit_yearly_total_vali_stoptake_min, net_profit_yearly_total_vali_stoptake_max,
                   share_pos_vali_stoptake_avg, share_pos_vali_stoptake_med, share_pos_vali_stoptake_min,
                   share_pos_vali_stoptake_max]]

# Results of the cross-validation are summarised in this dataframe. The values expressed are for all fields:
# average, median, minimum and maximum across the 5 folds performed.
summary_df_row = pd.DataFrame(summary_df_row, columns=summary_df_cols)


########################################################################################################################
######################################################## Plots #########################################################
########################################################################################################################

max_index = int(1/number_of_splits*len(Returns))
trades_matrix = np.zeros((len(trades_test), max_index))
trades_matrix_cum = np.copy(trades_matrix)

for i in range(len(trades_test.keys())):
    key = list(trades_test.keys())[i]
    for j in range(np.size(trades_test[key])):
        if np.size(trades_test[key]) == 1:
            index_t = int(trades_index[key])
        else:
            index_t = trades_index[key][j]
            trades_matrix[i, index_t] = trades_test[key][j]

for i in range(len(trades_test)):
    cumsum = np.cumsum(trades_matrix[i, :])
    trades_matrix_cum[i, :] = cumsum

avg_trades = np.zeros(max_index)
med_trades = np.zeros(max_index)
for i in range(max_index):
    avg_trades[i] = sum(trades_matrix_cum[:, i])/len(trades_test)
    if np.sum(trades_matrix_cum[trades_matrix_cum[:, i] != 0, i]) != 0:
        med_trades[i] = np.median(trades_matrix_cum[trades_matrix_cum[:, i] != 0, i])

plt.figure(figsize=(10, 6))
for i in range(len(trades_test.keys())):
    key = list(trades_test.keys())[i]
    plt.plot(trades_matrix_cum[i, :], label=f'Test {i}', linewidth=(i+1)*0.3)
    if np.size(trades_index[key]) == 1:
        plt.scatter(trades_index[key], trades_matrix_cum[i, int(trades_index[key])], color='k', )
    else:
        plt.scatter(trades_index[key], trades_matrix_cum[i, trades_index[key]], color='k', )

plt.plot(avg_trades, label=f'Average of Tests', linestyle='--', linewidth=2.5)
plt.plot(med_trades, label=f'Median of Tests', linestyle='--', linewidth=2.5)

plt.axhline(y=0, linestyle='--', color='black')
plt.legend()
plt.xlabel('Test Period Index after First Trade')
plt.ylabel('Cumulative Returns over Test Period [$]')
plt.title(f'Cumulative Returns over Test Period for the Strategy, no stop-loss')

# Plot 4:
fig, ax1 = plt.subplots()
ax1.plot(trades_matrix_cum[:, -1], linewidth=2.5, color='g')
ax1.set_xlabel('test index')
ax1.set_ylabel('Cumulative Returns at the end of test period [$]', color='g')
ax2 = ax1.twinx()
ax2.plot(std_test[list(trades_test.keys())], label='std', color='b', linestyle='--')
ax2.set_ylabel('std of test sample', color='b')
plt.title('Cumulative Returns at the end of test period [$] and std')
plt.show()
