import calendar
import numpy as np
from decimal import Decimal, ROUND_HALF_UP
from scipy.stats import norm
import decimal
import matplotlib.pyplot as plt
import xlwings as xw
import pandas as pd
import arch

########################################################################################################################
############################################### General Use ############################################################
########################################################################################################################

def EWMA_back(series, alpha):
    series_ewma = np.zeros(len(series))
    series_ewma[-1] = series[-1]
    for i in range(1, len(series)):
        series_ewma[-i - 1] = alpha * series[-i - 1] + (1 - alpha) * series_ewma[-i]
    return series_ewma

def CrackDataImport(max_holding_period, plot_ts, path):
    # Introducing additional 1 day buffer:
    max_holding_period = max_holding_period + 1
    # Importing the Price data from Excel:
    #SwapsTable = pd.read_excel('F:/TTF/Briac Thesis/prices_fwd.xlsx')
    SwapsTable = pd.read_excel(path)
    # Focusing on the Naphtha for now:
    NaphthaData = SwapsTable[SwapsTable['series_name'] == 'Naphtha CIF NWE Cargo']
    # Focusing on the Naphtha Crack for now:
    NaphthaDataCrack = NaphthaData[NaphthaData['series_type'] == 'Crack_Brent']
    # Focusing on the Naphtha Crack Front for now:
    NaphthaDataCrackFront = NaphthaDataCrack[NaphthaDataCrack['contract_month_rolling'] == 1]
    NaphthaDataCrackFront = NaphthaDataCrackFront.sort_values(by='eval_date')
    NaphthaDataCrackFront = NaphthaDataCrackFront.reset_index(drop=True)
    NaphthaDataCrackM2 = NaphthaDataCrack[NaphthaDataCrack['contract_month_rolling'] == 2]
    NaphthaDataCrackM2 = NaphthaDataCrackM2.sort_values(by='eval_date')
    NaphthaDataCrackM2 = NaphthaDataCrackM2.reset_index(drop=True)

    # Convert to string to then convert to decimal.
    NaphthaDataCrackFront['value'] = NaphthaDataCrackFront['value'].astype(str)
    DailyValues = NaphthaDataCrackFront['value'].reset_index(drop=True).values
    DailyValues = DailyValues.astype(float)
    DailyValues_dates = NaphthaDataCrackFront['eval_date'].reset_index(drop=True).values

    NaphthaDataCrackM2['value'] = NaphthaDataCrackM2['value'].astype(str)
    DailyValuesM2 = NaphthaDataCrackM2['value'].reset_index(drop=True).values
    DailyValuesM2 = DailyValuesM2.astype(float)
    DailyValues_dates_M2 = NaphthaDataCrackM2['eval_date'].reset_index(drop=True).values

    Contracts = pd.DataFrame({'eval_date': DailyValues_dates, 'M1': DailyValues})
    ContractsM2 = pd.DataFrame({'eval_date': DailyValues_dates_M2, 'M2': DailyValuesM2})

    merged_df = pd.concat([Contracts['eval_date'], ContractsM2['eval_date']])
    unique_values = merged_df.drop_duplicates(keep=False)
    ContractsM2 = ContractsM2[ContractsM2['eval_date'] != unique_values.values[0]]
    ContractsM1M2 = pd.merge(Contracts, ContractsM2, on='eval_date', how='inner')

    years = [dt.year for dt in ContractsM2['eval_date']]
    months = [dt.month for dt in ContractsM2['eval_date']]
    dates = pd.DataFrame({'years': years, 'months': months})
    dates_no_duplicates = dates.drop_duplicates().reset_index()

    roll_array = np.zeros(len(dates))
    for i in range(len(dates_no_duplicates)):
        year_month = dates.loc[(dates['years'] == dates_no_duplicates.loc[i, 'years']) & (dates['months'] == dates_no_duplicates.loc[i, 'months']), :]
        num_workdays = workdays_in_month(dates_no_duplicates.loc[i, 'years'], dates_no_duplicates.loc[i, 'months'])
        if i == 0:
            start_index_delete = year_month.index[0]
            end_index_delete = year_month.index[-1]

        if len(year_month) <= (num_workdays - max_holding_period):
            start_index = year_month.index[0]
            end_index = year_month.index[-1]
        else:
            start_index = year_month.index[0]
            end_index = year_month.index[num_workdays - max_holding_period]
        roll_array[start_index:end_index+1] = 1

    roll_array = roll_array.astype(int)
    ContractsM1M2['roll'] = roll_array
    ContractsM1M2 = ContractsM1M2.drop(ContractsM1M2.index[(start_index_delete):(end_index_delete+1)]).reset_index()

    sel_price = np.zeros(len(ContractsM1M2))
    for i in range(len(ContractsM1M2)):
        if ContractsM1M2.loc[i, 'roll'] == 1:
            sel_price[i] = ContractsM1M2.loc[i, 'M1']
        else:
            sel_price[i] = ContractsM1M2.loc[i, 'M2']

    ContractsM1M2['sel_price'] = sel_price

    if plot_ts == 1:
        plt.figure(figsize=(10, 6))
        plt.plot(ContractsM1M2['eval_date'], ContractsM1M2['sel_price'], label='early roll', marker='o', markersize=2)
        plt.plot(ContractsM1M2['eval_date'], ContractsM1M2['M1'], label='M1', marker='o', markersize=2)
        plt.xlabel('eval_date')
        plt.ylabel('Price [$]')
        plt.legend()

    DailyValues = ContractsM1M2['sel_price']
    DailyValues_dates = ContractsM1M2['eval_date']

    # Converting the strings to decimal.
    DailyValues = np.array([decimal.Decimal(elem) for elem in DailyValues])
    DailyValues = np.array([val.quantize(Decimal('0.001'), rounding=ROUND_HALF_UP) for val in DailyValues])

    DailyChange = DailyValues[1:] - DailyValues[:-1]
    DailyChange_dates = DailyValues_dates[1:]

    return DailyValues, DailyValues_dates, DailyChange, DailyChange_dates

def workdays_in_month(year, month): # Checked
    # Calculate the number of workdays in the specified month and year
    last_day = calendar.monthrange(year, month)[1]
    start_date = f'{year}-{month:02d}-01'  # Format month with leading zero if necessary
    end_date = f'{year}-{month:02d}-{last_day}'  # Format month with leading zero if necessary
    workdays = np.busday_count(start_date, end_date)
    return workdays

def EWMA(series, alpha):
    series_ewma = np.zeros(len(series))
    series_ewma[0] = series[0]
    for i in range(1, len(series)):
        series_ewma[i] = alpha * series[i] + (1 - alpha) * series_ewma[i - 1]
    return series_ewma

def RollingSmoothedStdV2(Returns, Returns_dates, window_std, alpha, plot_scaling):

    Nap_Std_Centered = np.zeros(len(range(int(window_std / 2), len(Returns) - int(window_std / 2))))
    Nap_Std_Backwards = np.zeros(len(range(int(window_std / 2), len(Returns))))
    Nap_Std_Dates_Centered = np.zeros(len(range(int(window_std / 2), len(Returns) - int(window_std / 2))), dtype=object)
    Nap_Std_Dates_Backwards = np.zeros(len(range(int(window_std / 2), len(Returns))), dtype=object)

    for i in range(int(window_std / 2), len(Returns) - int(window_std / 2)):
        # std formed with (window_std + 1) datapoints, window_std/2 before and window_std/2 after i.
        Nap_Std_Centered[i - int(window_std / 2)] = np.std(Returns[i - int(window_std / 2):i + int(window_std / 2) + 1])
        Nap_Std_Dates_Centered[i - int(window_std / 2)] = Returns_dates[i]

    for i in range(int(window_std / 2), len(Returns)):
        Nap_Std_Backwards[i - int(window_std / 2)] = np.std(Returns[i - int(window_std / 2):i])
        Nap_Std_Dates_Backwards[i - int(window_std / 2)] = Returns_dates[i]

    Nap_Std_Dates_Centered = pd.to_datetime(Nap_Std_Dates_Centered)
    Nap_Std_Dates_Backwards = pd.to_datetime(Nap_Std_Dates_Backwards)

    Nap_Std_Centered_EWMA = EWMA(Nap_Std_Centered, alpha)
    Nap_Std_Centered_EWMA_back = EWMA_back(Nap_Std_Centered, alpha)
    Nap_Std_Centered_EWMA_cent = (Nap_Std_Centered_EWMA + Nap_Std_Centered_EWMA_back) / 2
    Nap_Std_Backwards_EWMA = EWMA(Nap_Std_Backwards, alpha)

    if plot_scaling == 1:
        plt.figure(figsize=(10, 6))
        plt.plot(Nap_Std_Dates_Centered, Nap_Std_Centered, label='Centered STD')
        plt.plot(Nap_Std_Dates_Centered, Nap_Std_Centered_EWMA, label='Centered STD EWMA classic from start')
        plt.plot(Nap_Std_Dates_Centered, Nap_Std_Centered_EWMA_back, label='Centered STD EWMA classic from end')
        plt.plot(Nap_Std_Dates_Centered, Nap_Std_Centered_EWMA_cent, label='Centered STD EWMA classic average')
        plt.plot(Nap_Std_Dates_Backwards, Nap_Std_Backwards, label='Backwards STD')
        plt.xlabel('Date')
        plt.ylabel('Vol [$]')
        plt.legend()

    return Nap_Std_Dates_Centered, Nap_Std_Dates_Backwards, Nap_Std_Centered, Nap_Std_Centered_EWMA_cent, Nap_Std_Backwards, Nap_Std_Backwards_EWMA

def excel_read_append(file_path, sheet_name, new_rows):

    existing_df = pd.read_excel(file_path, sheet_name=sheet_name)

    if len(existing_df) != 0:
        existing_df = existing_df.iloc[:, 1:]
        new_df = pd.concat([existing_df, new_rows], ignore_index=True)
    else:
        new_df = new_rows

    app = xw.App(visible=True)
    wb = xw.Book(file_path)
    sheet = wb.sheets[sheet_name]
    sheet.clear()
    sheet.range('A1').value = new_df

    wb.save()
    wb.close()
    app.quit()

def scaling_objective_fun(Returns, Returns_dates, window_std, alpha_EWMA, series_choice):

    Std_Dates_Centered, Std_Dates_Backwards, Std_Centered, Std_Centered_EWMA_centered, Std_Backwards, Std_Backwards_EWMA = RollingSmoothedStdV2(Returns, Returns_dates, window_std, alpha_EWMA, 0)

    output_series = np.zeros(len(Returns))

    # Nap_Std_Centered_full_temp = Nap_Std_Centered_EWMA_cent_full.copy()
    if series_choice == 'std_centered_EWMA_centered':
        chosen_series = Std_Centered_EWMA_centered
        output_series[int(window_std / 2):-int(window_std / 2)] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
        output_series[-int(window_std / 2):] = chosen_series[-1]
    elif series_choice == 'std_backwards_EWMA':
        chosen_series = Std_Backwards_EWMA
        output_series[int(window_std / 2):] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
    elif series_choice == 'std_backwards':
        chosen_series = Std_Backwards
        output_series[int(window_std / 2):] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
    return output_series

def scaling_cali_fun(to_be_scaled_series, scaling_series):
    scaled_series = to_be_scaled_series / scaling_series
    std_cali = np.std(to_be_scaled_series)
    std_cali_scaled = np.std(scaled_series)
    scaled_series = scaled_series / std_cali_scaled * std_cali
    scaled_series = np.array([x.quantize(Decimal('0.001'), rounding=ROUND_HALF_UP) for x in scaled_series])
    return scaled_series, std_cali, std_cali_scaled

def scaling_test_fun(to_be_scaled_series, scaling_series, std_cali, std_cali_scaled):
    scaled_series = to_be_scaled_series / scaling_series
    scaled_series = scaled_series / std_cali_scaled * std_cali
    scaled_series = np.array([x.quantize(Decimal('0.001'), rounding=ROUND_HALF_UP) for x in scaled_series])
    return scaled_series

def Strategy_1_Fun(ChangeMatrix_input):  # Checked
    ChangeMatrix_filtered = ChangeMatrix_input.copy()
    # Description: Just enter the position one day after the trigger.
    return ChangeMatrix_filtered

def Strategy_2_Fun(ChangeMatrix_input):  # Checked
    ChangeMatrix_filtered = ChangeMatrix_input.copy()
    # Description:
    # Applies cumulative returns computation but introducing a 1-Day delay in the trade entry.
    # Implementing a stop-loss/take-profit modification of the change matrix if requested.
    # Computes expected returns and other metrics using SignificativeReversion.

    # Delaying ChangeMatrix to account for the wait until t+1, t+1 variation taken out of the cumulative variation:
    ChangeMatrix_filtered[:, 1:] = ChangeMatrix_filtered[:, 1:] - ChangeMatrix_filtered[:, 1][:, np.newaxis]
    return ChangeMatrix_filtered

def Strategy_3_Fun(ChangeMatrix_input):
    ChangeMatrix_filtered = ChangeMatrix_input.copy()
    # Description:
    # Only keep non-reverting values at t+1.
    # Applies cumulative returns computation but introducing a 1-Day delay in the trade entry.
    # Implementing a stop-loss/take-profit modification of the change matrix if requested.
    # Computes expected returns and other metrics using SignificativeReversion.

    # Apply additional filter: continuation at t+1
    ChangeMatrix_filtered = ChangeMatrix_filtered[ChangeMatrix_filtered[:, 1] < 0, :]
    # Delaying ChangeMatrix to account for the wait until t+1, t+1 variation taken out of the cumulative variation:
    ChangeMatrix_filtered[:, 1:] = ChangeMatrix_filtered[:, 1:] - ChangeMatrix_filtered[:, 1][:, np.newaxis]
    return ChangeMatrix_filtered

def FDR_control(p_values_matrix, alpha_FDR, plot_activation):

    vector_p_value = p_values_matrix.flatten()
    sorted_indices = np.argsort(vector_p_value)
    vector_p_value = vector_p_value[sorted_indices]

    Benjamini_val = np.empty(len(vector_p_value))
    n_tests = len(vector_p_value)

    for i in range(len(vector_p_value)):
        Benjamini_val[i] = alpha_FDR*(i+1)/n_tests

    if np.all(np.array([vector_p_value <= Benjamini_val]) == False):
        alpha_FDR_new = 0
    else:
        alpha_FDR_new = np.max(vector_p_value[vector_p_value <= Benjamini_val])
        index = np.argwhere(vector_p_value == alpha_FDR_new)[0][0]

        if plot_activation == 1:
            plt.figure()
            plt.plot(vector_p_value, label='Sorted vector of p-values')
            plt.scatter(range(len(vector_p_value)), vector_p_value)
            plt.plot(Benjamini_val, label='Benjamini-Hochberg criterion')
            plt.scatter(index, alpha_FDR_new, color='red', label='Largest criterion')
            plt.legend()
            plt.xlabel('index of p_values')
            plt.ylabel('value')

    return alpha_FDR_new, vector_p_value

def build_delta_matrix(daily_delta, daily_delta_scaled, max_holding_period, scaling):

    rows = len(daily_delta) - max_holding_period
    cols = max_holding_period + 1
    delta_matrix = np.zeros((rows, cols), dtype=object)

    for k in range(rows):
        if scaling == 1:
            delta_matrix[k, 0] = daily_delta_scaled[k]
        else:
            delta_matrix[k, 0] = daily_delta[k]

        for l in range(2, max_holding_period + 2):
            delta_matrix[k, l - 1] = np.sum(daily_delta[k + 1:k + l])

    # Making the first move always positive, the following cumulative sums positive if reverting, negative if confirming.
    delta_matrix_raw = delta_matrix.copy()
    delta_matrix_raw[:, 0] = daily_delta[:-max_holding_period]
    negative_indices = delta_matrix[:, 0] < 0
    positive_indices = delta_matrix[:, 0] >= 0
    delta_matrix[negative_indices, 0] *= -1
    delta_matrix[positive_indices, 1:] *= -1

    return delta_matrix, delta_matrix_raw

def SignificativeReversion(ChangeMatrix_pop, ChangeMatrix_sample, reversion_period, tail): # checked
    # Description:
    # Computes expected returns.
    # Computes statistical test for expected returns.
    # Perform Statistical Test, Z-test because pop std is known and n is large:
    sample_mean = float(np.average(ChangeMatrix_sample[:, reversion_period]))  # Sample mean
    population_mean = 0  # Population mean
    population_std = float(np.std(ChangeMatrix_pop[:, reversion_period]))  # Population standard deviation
    sample_size = len(ChangeMatrix_sample[:, reversion_period])  # Sample size

    p_value, z_statistic = compute_z_test_p_value(sample_mean, population_mean, population_std, sample_size, tail)

    p_value = round(p_value, 10)

    return p_value

def compute_z_test_p_value(sample_mean, population_mean, population_std, sample_size, tail):

    # Calculate the standard error of the mean
    std_error = population_std / np.sqrt(sample_size)
    # Calculate the z-statistic
    z_statistic = (sample_mean - population_mean) / std_error

    # Calculate the p-value
    if tail == 'two-tailed':
        p_value = 2 * (1 - norm.cdf(abs(z_statistic)))
    elif tail == 'one-tailed-lower':
        p_value = norm.cdf(z_statistic)
    elif tail == 'one-tailed-upper':
        p_value = 1 - norm.cdf(z_statistic)
    else:
        raise ValueError("tail must be 'two-tailed', 'one-tailed-lower', or 'one-tailed-upper'")
    return p_value, z_statistic

def Rolling_Garch(series, pred_horizon, p_param, q_param):
    Garch_Implied_Vol = np.zeros(len(series))
    for i in range(1, len(series)):
        train = series[:i + 1]
        model = arch.arch_model(train, p=p_param, q=q_param, vol='GARCH', dist='Normal', rescale=True)
        model_fit = model.fit(disp='off')
        scale = model_fit.scale
        pred = model_fit.forecast(horizon=pred_horizon)
        Garch_Implied_Vol[i] = np.sqrt(pred.variance.values[-1, :] / np.power(scale, 2))
        print(f'GARCH {round(i / (len(series) - 1) * 100, 2)}%')
    return Garch_Implied_Vol