import sys
sys.path.append()
import Functions
from sklearn.neighbors import KernelDensity
from scipy.stats import norm
from sklearn.model_selection import GridSearchCV
from scipy.stats import skew, kurtosis
from statsmodels.tsa.stattools import adfuller
import statsmodels.api as sm
from scipy.stats import skew
from scipy.stats import johnsonsu
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from mpl_toolkits.axes_grid1.inset_locator import mark_inset


def scaling_objective_fun(Returns, Returns_dates, window_std, alpha_EWMA, series_choice):

    Std_Dates_Centered, Std_Dates_Backwards, Std_Centered, Std_Centered_EWMA_centered, Std_Backwards, Std_Backwards_EWMA = Functions.RollingSmoothedStdV2(Returns, Returns_dates, window_std, alpha_EWMA, 0)

    output_series = np.zeros(len(Returns))

    # Nap_Std_Centered_full_temp = Nap_Std_Centered_EWMA_cent_full.copy()
    if series_choice == 'std_centered_EWMA_centered':
        chosen_series = Std_Centered_EWMA_centered
        output_series[int(window_std / 2):-int(window_std / 2)] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
        output_series[-int(window_std / 2):] = chosen_series[-1]
    elif series_choice == 'std_backwards_EWMA':
        chosen_series = Std_Backwards_EWMA
        output_series[int(window_std / 2):] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
    elif series_choice == 'std_backwards':
        chosen_series = Std_Backwards
        output_series[int(window_std / 2):] = chosen_series
        output_series[0:int(window_std / 2)] = chosen_series[0]
    return output_series

def scaling_cali_fun(to_be_scaled_series, scaling_series):
    scaled_series = to_be_scaled_series / scaling_series
    std_cali = np.std(to_be_scaled_series)
    std_cali_scaled = np.std(scaled_series)
    scaled_series = scaled_series / std_cali_scaled * std_cali
    return scaled_series, std_cali, std_cali_scaled

def non_param_estimation(data, n_x, bandwidth, scaling_type):

    ########################################################################################################################
    ########################################## Non parametric estimation: ##################################################
    ########################################################################################################################

    kernel_type = 'gaussian'
    kde = KernelDensity(kernel=kernel_type, bandwidth=bandwidth).fit(data)
    x_d = np.linspace(min(data), max(data), 1000)
    log_density = kde.score_samples(x_d)
    density = np.exp(log_density)

    tot_percentile = 99
    lower = (100-tot_percentile)/2
    upper = 100 - lower
    lower_percentile = np.percentile(data, lower)
    upper_percentile = np.percentile(data, upper)

    plt.figure()
    plt.hist(data, density=True, bins=int(np.sqrt(len(data))), color='gray', alpha=0.5)
    plt.plot(x_d, density, label='Gaussian kernel density', color='k')
    plt.axvline(lower_percentile, color='r', linestyle='dashed', linewidth=2,  label=f'percentiles: {lower}%, {upper}%')
    plt.axvline(upper_percentile, color='r', linestyle='dashed', linewidth=2)
    plt.axvline(0, color='black', linestyle='-', linewidth=0.5)
    plt.axhline(0, color='black', linestyle='-', linewidth=0.5)
    plt.legend(loc='upper right')
    plt.grid()
    plt.ylim([0, 1])
    plt.title(f'standardisation: {scaling_type}')
    plt.xlabel('Standardised delta [$/bbl]')
    plt.ylabel('Density [-]')


    ########################################################################################################################
    ########################################## Drift estimation: ###########################################################
    ########################################################################################################################

    x = np.linspace(min(data), max(data), n_x)
    mu_x = np.zeros(len(x))
    mu2_x = np.zeros(len(x))
    mu3_x = np.zeros(len(x))
    sig_x = np.zeros(len(x))
    sig2_x = np.zeros(len(x))
    sig3_x = np.zeros(len(x))
    sig_x_var = np.zeros(len(x))
    sig2_x_var = np.zeros(len(x))
    sig3_x_var = np.zeros(len(x))

    for i in range(len(x)):

        kern_1 = np.zeros(len(data) - 1)
        diff_1 = np.zeros(len(data) - 1)
        diff_var_1 = np.zeros(len(data) - 1)
        kern_2 = np.zeros(len(data) - 2)
        diff_2 = np.zeros(len(data) - 2)
        diff_var_2 = np.zeros(len(data) - 2)
        kern_3 = np.zeros(len(data) - 3)
        diff_3 = np.zeros(len(data) - 3)
        diff_var_3 = np.zeros(len(data) - 3)

        # num_vals = np.zeros(len(data) - 1)
        # denom_vals = np.zeros(len(data) - 1)
        # num_vals_2_1 = np.zeros(len(data) - 1)
        # denom_vals_2_1 = np.zeros(len(data) - 1)
        # num_vals_2_2 = np.zeros(len(data) - 2)
        # denom_vals_2_2 = np.zeros(len(data) - 2)
        # num_vals_3_1 = np.zeros(len(data) - 1)
        # denom_vals_3_1 = np.zeros(len(data) - 1)
        # num_vals_3_2 = np.zeros(len(data) - 2)
        # denom_vals_3_2 = np.zeros(len(data) - 2)

        for j in range(len(data) - 1):

            diff_1[j] = data[j + 1] - data[j]
            kern_1[j] = norm.pdf((x[i] - data[j]) / bandwidth)

            # num_vals[j] = (data[j + 1] - data[j]) * norm.pdf((x[i] - data[j]) / bandwidth)
            # denom_vals[j] = norm.pdf((x[i] - data[j]) / bandwidth)
            #
            # num_vals_2_1[j] = (data[j + 1] - data[j]) * norm.pdf((x[i] - data[j]) / bandwidth)
            # denom_vals_2_1[j] = norm.pdf((x[i] - data[j]) / bandwidth)

        for j in range(len(data) - 2):

            diff_2[j] = data[j + 2] - data[j]
            kern_2[j] = norm.pdf((x[i] - data[j]) / bandwidth)

            # num_vals_2_2[j] = (data[j + 2] - data[j]) * norm.pdf((x[i] - data[j]) / bandwidth)
            # denom_vals_2_2[j] = norm.pdf((x[i] - data[j]) / bandwidth)

        for j in range(len(data)-3):

            diff_3[j] = data[j + 3] - data[j]
            kern_3[j] = norm.pdf((x[i] - data[j]) / bandwidth)

            # num_vals_3_2[j] = (data[j + 3] - data[j]) * norm.pdf((x[i] - data[j]) / bandwidth)
            # denom_vals_3_2[j] = norm.pdf((x[i] - data[j]) / bandwidth)

        mu_x[i] = np.sum(diff_1*kern_1)/np.sum(kern_1)
        mu2_x[i] = (1/2)*(4*np.sum(diff_1*kern_1)/np.sum(kern_1) - np.sum(diff_2*kern_2)/np.sum(kern_2))
        mu3_x[i] = (1/6)*(18*np.sum(diff_1*kern_1)/np.sum(kern_1) - 9*np.sum(diff_2*kern_2)/np.sum(kern_2) + 2*np.sum(diff_3*kern_3)/np.sum(kern_3))

        # mu_x[i] = np.sum(num_vals) / np.sum(denom_vals)
        # mu2_x[i] = 0.5*(4*np.sum(num_vals_2_1)/np.sum(denom_vals_2_1) - np.sum(num_vals_2_2)/np.sum(denom_vals_2_2))
        # mu3_x[i] = (1/6)*(18 * np.sum(num_vals_2_1)/np.sum(denom_vals_2_1) - 9 * np.sum(num_vals_2_2)/np.sum(denom_vals_2_2) + 2 * np.sum(num_vals_3_2)/np.sum(denom_vals_3_2))


        for j in range(len(data) - 1):
            diff_var_1[j] = data[j + 1] - data[j] - mu_x[i]

        for j in range(len(data) - 2):
            diff_var_2[j] = data[j + 2] - data[j] - mu2_x[i]

        for j in range(len(data)-3):
            diff_var_3[j] = data[j + 3] - data[j] - mu3_x[i]


        sig_x[i] = np.sqrt(np.sum(diff_1**2 * kern_1)/np.sum(kern_1))
        sig2_x[i] = np.sqrt((1/2)*(4*np.sum(diff_1**2 * kern_1)/np.sum(kern_1) - np.sum(diff_2**2 * kern_2)/np.sum(kern_2)))
        sig3_x[i] = np.sqrt((1/6)*(18*np.sum(diff_1**2 * kern_1)/np.sum(kern_1) - 9*np.sum(diff_2**2 * kern_2)/np.sum(kern_2) + 2*np.sum(diff_3**2 * kern_3)/np.sum(kern_3)))

        sig_x_var[i] = np.sqrt(np.sum(diff_var_1**2 * kern_1)/np.sum(kern_1))
        sig2_x_var[i] = np.sqrt((1/2)*(4*np.sum(diff_var_1**2 * kern_1)/np.sum(kern_1) - np.sum(diff_var_2**2 * kern_2)/np.sum(kern_2)))
        sig3_x_var[i] = np.sqrt((1/6)*(18*np.sum(diff_var_1**2 * kern_1)/np.sum(kern_1) - 9*np.sum(diff_var_2**2 * kern_2)/np.sum(kern_2) + 2*np.sum(diff_var_3**2 * kern_3)/np.sum(kern_3)))


    fig, ax1 = plt.subplots()
    ax1.plot(x, mu_x, label='1st order', color='k', linestyle=':')
    ax1.plot(x, mu2_x, label='2nd order', color='k', linestyle='--')
    ax1.plot(x, mu3_x, label='3rd order', color='k', linestyle='-')
    ax1.axvline(lower_percentile, color='r', linestyle='dashed', linewidth=2, label=f'percentiles: {lower}%, {upper}%')
    ax1.axvline(upper_percentile, color='r', linestyle='dashed', linewidth=2)
    ax1.axvline(0, color='black', linestyle='-', linewidth=0.5)
    ax1.axhline(0, color='black', linestyle='-', linewidth=0.5)
    ax1.set_xlabel('Daily standardised delta level [$\$/bbl$]')
    ax1.set_ylabel(r'Drift term [$\$/bbl \cdot day^{-1}$]')
#    ax1.legend(loc=(0.8, 0.8))
    ax1.legend(loc=(0.53, 0.1))
    ax1.set_xlim(min(x)*0.8, 2.4)
    #ax1.set_title(f'c = {c}, standardisation: {scaling_type}')
    ax2 = ax1.twinx()
    ax2.hist(data, bins=int(np.sqrt(len(data))), alpha=0.5, color='gray')
    ax2.set_ylabel('Frequency')
    ax2.set_ylim(0, 11)
    ax2.legend(['Histogram of standardised delta'], loc='lower right')
    plt.show()


    ########################################################################################################################
    ########################################## Diffusion estimation: #######################################################
    ########################################################################################################################
    #
    # x = np.linspace(min(data), max(data), n_x)
    # sig_x = np.zeros(len(x))
    # sig2_x = np.zeros(len(x))
    # sig3_x = np.zeros(len(x))
    #
    # for i in range(len(x)):
    #     num_vals = np.zeros(len(data) - 1)
    #     num_vals_var = np.zeros(len(data) - 1)
    #     denom_vals = np.zeros(len(data) - 1)
    #     num_vals_2_1 = np.zeros(len(data) - 1)
    #     denom_vals_2_1 = np.zeros(len(data) - 1)
    #     num_vals_2_2 = np.zeros(len(data) - 2)
    #     denom_vals_2_2 = np.zeros(len(data) - 2)
    #     num_vals_3_2 = np.zeros(len(data) - 2)
    #     denom_vals_3_2 = np.zeros(len(data) - 2)
    #
    #     for j in range(len(data)-1):
    #
    #         num_vals[j] = (data[j+1] - data[j] - mu_x[i])**2 * norm.pdf((x[i] - data[j]) / bandwidth)
    #         #num_vals[j] = (data[j+1] - data[j])**2 * norm.pdf((x[i] - data[j]) / bandwidth)
    #         denom_vals[j] = norm.pdf((x[i] - data[j]) / bandwidth)
    #
    #         num_vals_2_1[j] = (data[j+1] - data[j] - mu_x[i])**2 * norm.pdf((x[i] - data[j]) / bandwidth)
    #         #num_vals_2_1[j] = (data[j+1] - data[j])**2 * norm.pdf((x[i] - data[j]) / bandwidth)
    #         denom_vals_2_1[j] = norm.pdf((x[i] - data[j]) / bandwidth)
    #
    #     for j in range(len(data) - 2):
    #
    #         #num_vals_2_2[j] = (data[j+2] - data[j] - mu2_x[i])**2 * norm.pdf((x[i] - data[j]) / bandwidth)
    #         num_vals_2_2[j] = (data[j+2] - data[j])**2 * norm.pdf((x[i] - data[j]) / bandwidth)
    #         denom_vals_2_2[j] = norm.pdf((x[i] - data[j]) / bandwidth)
    #
    #     for j in range(len(data) - 3):
    #
    #         num_vals_3_2 = (data[j+3] - data[j])**2 * norm.pdf((x[i] - data[j]) / bandwidth)
    #         denom_vals_3_2 = norm.pdf((x[i] - data[j]) / bandwidth)
    #
    #     sig_x[i] = (np.sum(num_vals)/np.sum(denom_vals))**(1/2)
    #     sig2_x[i] = ((1/2)*(4*np.sum(num_vals_2_1)/np.sum(denom_vals_2_1)) - (np.sum(num_vals_2_2) / np.sum(denom_vals_2_2)))**(1/2)
    #     sig3_x[i] = ((1/6)*(18*(np.sum(num_vals_2_1)/np.sum(denom_vals_2_1)) - 9*(np.sum(num_vals_2_2)/np.sum(denom_vals_2_2)) + 2*(np.sum(num_vals_3_2)/np.sum(denom_vals_3_2))))**(1 / 2)


    fig, ax1 = plt.subplots()
    # ax1.plot(x, sig_x, label='1st order exp', color='#3776ab', linestyle=':')
    # ax1.plot(x, sig2_x, label='2nd order exp', color='#3776ab', linestyle='--')
    # ax1.plot(x, sig3_x, label='3rd order exp', color='#3776ab', linestyle='-')
    ax1.plot(x, sig_x_var, label='1st order var', color='k', linestyle=':')
    ax1.plot(x, sig2_x_var, label='2nd order var', color='k', linestyle='--')
    ax1.plot(x, sig3_x_var, label='3rd order var', color='k', linestyle='-')
    ax1.axvline(lower_percentile, color='r', linestyle='dashed', linewidth=2, label=f'percentiles: {lower}%, {upper}%')
    ax1.axvline(upper_percentile, color='r', linestyle='dashed', linewidth=2)
    ax1.axvline(0, color='black', linestyle='-', linewidth=0.5)
    ax1.axhline(0, color='black', linestyle='-', linewidth=0.5)
    ax1.legend()
    ax1.set_xlim(min(x)*0.8, 2.4)
    ax1.set_xlabel('Daily Delta Level [$\$/bbl$]')
    ax1.set_ylabel(r'Diffusion term [$\$/bbl \cdot day^{-1/2}$]')
    #ax1.set_title(f'c = {c}, standardisation: {scaling_type}')
    ax2 = ax1.twinx()
    ax2.hist(data, bins=int(np.sqrt(len(data))), alpha=0.5, color='gray')
    ax2.set_ylabel('Frequency')
    ax2.set_ylim(0, 11)  # Adjust the ylim to fit the histogram nicely
    #ax1.legend(loc=(0.0075, 0.71))
    ax1.legend(loc=(0.015, 0.65))
    ax2.legend(['Histogram of standardised delta'], loc='upper left')
    plt.show()

    return density, mu_x, sig_x, sig_x_var, mu2_x, sig2_x, sig2_x_var, mu3_x, sig3_x, sig3_x_var, x, kde

def build_delta_matrix(row_means_matrix):

    rows = np.shape(row_means_matrix)[0]
    cols = np.shape(row_means_matrix)[1]
    delta_matrix = np.zeros((rows, cols))

    delta_matrix[:, 0] = row_means_matrix[:, 0]

    for k in range(rows):
        for l in range(2, cols + 1):
            delta_matrix[k, l - 1] = np.sum(row_means_matrix[k, 1:l])

    # Making the first move always positive, the following cumulative sums positive if reverting, negative if confirming.
    positive_indices = delta_matrix[:, 0] >= 0
    delta_matrix[positive_indices, 1:] *= -1

    return delta_matrix

def recursive_arithmetic_brownian(n_days, start_del, mu_x, sig_x, t_step, x):

    delta_vec = np.zeros(n_days + 1)
    delta_vec[0] = start_del

    for i in range(n_days):
        index_closest = np.abs(x - delta_vec[i]).argmin()
        delta_vec[i + 1] = delta_vec[i] + mu_x[index_closest] * t_step + sig_x[index_closest] * np.sqrt(t_step) * np.random.randn()

    return delta_vec

def compute_statistics(data):

    statistics = {}

    # Convert data to a numpy array
    data = np.asarray(data)

    # Compute Mean
    statistics['Mean'] = np.mean(data)

    # Compute Standard Deviation
    statistics['Standard dev.'] = np.std(data)

    # Compute Minimum
    statistics['Minimum'] = np.min(data)

    # Compute Maximum
    statistics['Maximum'] = np.max(data)

    # Compute Skewness
    statistics['Skewness'] = skew(data)

    # Compute Kurtosis
    statistics['Kurtosis'] = kurtosis(data)

    # Compute Dickey-Fuller statistic
    adf_result = adfuller(data, regression='n')
    statistics['DF stat'] = adf_result[0]  # The test statistic value
    print(adf_result[1])
    statistics['p_value'] = adf_result[1]

    return statistics

def linear_OLS(X, Y):

    X = sm.add_constant(X)  # Add a constant term for the intercept

    # Fit the OLS model
    model_main = sm.OLS(Y, X).fit()
    y_pred = model_main.predict(X)

    # Plotting the residuals
    residuals = Y - model_main.predict(X)

    # # Plot the data and the fitted line with confidence intervals
    # plt.figure(figsize=(10, 6))
    # plt.scatter(X[:, 1], Y, marker='+', alpha=0.9, s=10)
    # plt.plot(X[:, 1], y_pred, color='red', label='OLS Regression Line')
    # plt.grid()
    # plt.xlabel('Standardised Delta [$/bbl]')
    # plt.ylabel('Delta [$/bbl]')
    # plt.legend()
    # plt.show()
    #
    # plt.figure(figsize=(10, 6))
    # plt.grid()
    # plt.scatter(X[:, 1], residuals, alpha=0.5)
    # plt.axhline(0, color='red', linestyle='--')
    # plt.xlabel('Standardised Delta [$/bbl]')
    # plt.ylabel('Residuals [$/bbl]')
    # plt.show()

    return model_main, residuals

def check_max_min_param(series, param, deviation):

    if np.max(series) >= 0:
        if param > (1 + deviation) * np.max(series):
            param = (1 + deviation) * np.max(series)

    elif np.max(series) < 0:
        if param > (1 - deviation) * np.max(series):
            param = (1 - deviation) * np.max(series)

    if np.min(series) >= 0:
        if param < (1 - deviation) * np.min(series):
            param = (1 - deviation) * np.min(series)

    elif np.min(series) < 0:
        if param < (1 + deviation) * np.min(series):
            param = (1 + deviation) * np.min(series)

    return param

def johnsonsu_descaling(val, model_main, model_a, model_loc, a_vec, b_vec, loc_vec, scale_vec, bucket_mid, deviation):

    val_x = np.array([[1, val]])

    a_param = model_a.predict(val_x)[0]
    a_param = check_max_min_param(a_vec[~np.isnan(a_vec)], a_param, deviation)

    b_param = np.interp(val, bucket_mid[~np.isnan(b_vec)], b_vec[~np.isnan(b_vec)])
    b_param = check_max_min_param(b_vec[~np.isnan(b_vec)], b_param, deviation)

    loc_param = model_loc.predict(val_x)[0]
    loc_param = check_max_min_param(loc_vec[~np.isnan(loc_vec)], loc_param, deviation)

    scale_param = np.interp(val, bucket_mid[~np.isnan(scale_vec)], scale_vec[~np.isnan(scale_vec)])
    scale_param = check_max_min_param(scale_vec[~np.isnan(scale_vec)], scale_param, deviation)

    johnsonsu_dist = johnsonsu(a_param, b_param, loc=loc_param, scale=scale_param)

    model_residuals = johnsonsu_dist.rvs()
    model_base = model_main.predict(val_x)

    descaled_val = model_base + model_residuals

    return descaled_val


########################################################################################################################
############################################### Select Parameters #1 ###################################################
########################################################################################################################

# Insert path to the Excel file containing the price data: NaphthaCrackPaper.xlsx
path = input("Input the path to the Excel file NaphthaCrackPaper.xlsx: ")

# Select the standardising series: 'backwards std 20 days', 'backwards std 60 days', 'GARCH(1,1)'
scaling_type = input("Select the standardising series: backwards std 20 days, backwards std 60 days, GARCH(1,1): ")

# Select the smoothing coefficient:
alpha_EWMA = float(input("Select the EWMA smoothing coefficient (base case is 0.2): "))


########################################################################################################################
############################################### Extract Deltas #########################################################
########################################################################################################################
Max_Reversion_Period = 5
plot_ts = 1
# Extract the deltas:
DailyValues, DailyValues_dates, Deltas, Deltas_dates = Functions.CrackDataImport(Max_Reversion_Period, plot_ts, path)

# Convert from Decimal to floats:
DailyValues = np.array([float(d) for d in DailyValues])
Deltas = np.array([float(d) for d in Deltas])

statistics = compute_statistics(Deltas)


########################################################################################################################
############################################### Scale and Smooth #######################################################
########################################################################################################################

if scaling_type == 'backwards std 60 days':
    days = 120
    # Get the scaling_series with corresponding smoothing:
    scaling_series = scaling_objective_fun(Deltas, Deltas_dates, days, alpha_EWMA, 'std_backwards_EWMA')

elif scaling_type == 'backwards std 20 days':
    days = 40
    # Get the scaling_series with corresponding smoothing:
    scaling_series = scaling_objective_fun(Deltas, Deltas_dates, days, alpha_EWMA, 'std_backwards_EWMA')

else:
    # Garch Parameters:
    pred_horizon = 1
    p_param = 1
    q_param = 1
    garch_stabilisation_delay = 200

    # Computing Garch:
    Garch_Implied_Vol_full = Functions.Rolling_Garch(Deltas, pred_horizon, p_param, q_param)
    Garch_Implied_Vol = Garch_Implied_Vol_full[garch_stabilisation_delay:]
    Garch_dates = Deltas_dates[garch_stabilisation_delay:]
    Deltas = Deltas[garch_stabilisation_delay:]
    scaling_series = Functions.EWMA(Garch_Implied_Vol, alpha_EWMA)

# Scale the deltas:
scaled_deltas, std_cali, std_cali_scaled = scaling_cali_fun(Deltas, scaling_series)

# Get basic stats
statistics = compute_statistics(scaled_deltas)

# Format the data for input:
data = scaled_deltas[:, np.newaxis]


########################################################################################################################
############################################### Select Parameters #2 ###################################################
########################################################################################################################

n_x = 200  # Number of points for the estimation of delta in the range min(delta), max(delta)
n_days = 5  # Number of days for the Monte Carlo Simulation.
t_step = 1  # timestep for the arithmetic brownian motion (keep a one for daily moves).
n_simu = 10000  # number of simulations for the montecarlo.


########################################################################################################################
################################### Fit non-parametric arithmetic Brownian Motion ######################################
########################################################################################################################

# Get the best bandwidth via cross-validation:
bandwidths = np.logspace(-1, 1, 20)
grid = GridSearchCV(KernelDensity(kernel='gaussian'), {'bandwidth': bandwidths}, cv=10)  # 10-fold cross-validation
grid.fit(data)
best_bandwidth = grid.best_params_['bandwidth']
c = round(best_bandwidth / np.std(data) * len(data) ** (1 / 5), 2)

# Fit non-parametric distribution:
density, mu_x, sig_x, sig_x_var, mu2_x, sig2_x, sig2_x_var, mu3_x, sig3_x, sig3_x_var, x, kde = non_param_estimation(data, n_x, best_bandwidth, scaling_type)

chosen_mu = mu3_x.copy()
chosen_sig = sig_x_var.copy()


########################################################################################################################
################################### Run Monte-Carlo Simulations  #######################################################
########################################################################################################################

# Choose threshold above which position is opened:
chosen_threshold = 1.6

# Generate n_simu values that are beyond the designated threshold:
I = 0
val_array_all = np.empty(0)
val_array = np.zeros(n_simu)

while I < n_simu:

    val = kde.sample()[0, 0]
    val_array_all = np.append(val_array_all, np.array(val))

    if (val >= chosen_threshold) | (val <= -chosen_threshold):
        val_array[I] = val
        I = I + 1
        print(f'Progress: {round(I/n_simu*100, 2)}%')


# For each, generate an arithmetic brownian using our parameters:
delta_matrix = np.zeros((len(val_array), n_days+1))
for i in range(len(val_array)):
    start_del = val_array[i]
    delta_vec = recursive_arithmetic_brownian(n_days, start_del, chosen_mu, chosen_sig, t_step, x)
    delta_matrix[i, :] = delta_vec

# Compute the PnL for the lossses
final_matrix = build_delta_matrix(delta_matrix)
final_matrix[:, 0] = 0

day5avg = np.zeros(len(final_matrix))
for i in range(len(final_matrix)):
    day5avg[i] = np.average(final_matrix[0:i+1, 4])

std_day5avg = np.zeros(len(day5avg))
for i in range(10, len(std_day5avg)-10):
    std_day5avg[i] = np.std(day5avg[i-10:i+10])

fig, ax1 = plt.subplots()
color = 'tab:blue'
ax1.set_xlabel('Number of Simulations')
ax1.set_ylabel('PnL [$]', color=color)
ax1.plot(day5avg, color=color)
ax1.tick_params(axis='y', labelcolor=color)
# Create a secondary y-axis
ax2 = ax1.twinx()  # instantiate a second y-axis that shares the same x-axis
color = 'tab:red'
ax2.set_ylabel('std 20 days centered', color=color)
ax2.plot(std_day5avg, color=color)
ax2.tick_params(axis='y', labelcolor=color)
# Title and show the plot
plt.title('Plot with Two Different Y-axes')
fig.tight_layout()  # adjust subplots to fit into figure area.
plt.show()

mean = np.mean(final_matrix, axis=0)
mean_scaled = mean.copy()
std = np.std(final_matrix, axis=0)
conf_interval = 1.96 * std / np.sqrt(final_matrix.shape[0])

plt.figure(figsize=(10, 6))
plt.plot(mean, label='Mean')
plt.fill_between(np.arange(mean.shape[0]), mean - conf_interval, mean + conf_interval, color='b', alpha=0.2, label='95% Confidence Interval')
plt.axhline(y=0.2, color='red', linestyle='dotted', linewidth=2, label=f'Trading costs without stop loss ($0.20)')
plt.title(f'Monte Carlo Simulation, {scaling_type} EWMA {alpha_EWMA}, Threshold = ${chosen_threshold}')
plt.ylim([-2, 2])
plt.grid()
plt.xlabel('Days after Trigger')
plt.ylabel('PnL of Reversion [$]')
plt.legend(loc='lower right')
plt.show()

plt.figure()
plt.hist(data, bins=int(np.sqrt(len(data))),  density=True, alpha=0.5, label='actuals')
plt.hist(val_array_all, bins=int(np.sqrt(len(val_array_all))), density=True, alpha=0.5, label='modelled')
plt.legend()
plt.xlabel('Delta')
plt.ylabel('Density')


########################################################################################################################
################################### Link Scaled and Unscaled Deltas  ###################################################
########################################################################################################################

model_main, residuals = linear_OLS(scaled_deltas, Deltas)

n_buckets = 20
pop_lim = 30
scaled_deltas_buckets = np.linspace(np.min(scaled_deltas), np.max(scaled_deltas), n_buckets)
bucket_pop = np.zeros((n_buckets-1))
bucket_mean = np.zeros((n_buckets-1))
bucket_std = np.zeros((n_buckets-1))
bucket_skew = np.zeros((n_buckets-1))
a_vec = np.zeros((n_buckets-1))
b_vec = np.zeros((n_buckets-1))
loc_vec = np.zeros((n_buckets-1))
scale_vec = np.zeros((n_buckets-1))
bucket_mid = (scaled_deltas_buckets[:-1] + scaled_deltas_buckets[1:])/2

for i in range(n_buckets-1):

    if i < n_buckets-2:
        residuals_bucket = residuals[(scaled_deltas >= scaled_deltas_buckets[i]) & (scaled_deltas < scaled_deltas_buckets[i + 1])]
    else:
        residuals_bucket = residuals[(scaled_deltas >= scaled_deltas_buckets[i])]

    bucket_pop[i] = len(residuals_bucket)
    bucket_mean[i] = np.mean(residuals_bucket)

    if len(residuals_bucket) < 2:
        bucket_std[i] = np.nan
        bucket_skew[i] = np.nan
    else:
        bucket_std[i] = np.std(residuals_bucket)
        bucket_skew[i] = skew(residuals_bucket)

    if bucket_pop[i] >= pop_lim:

        x_vals = np.linspace(np.min(residuals_bucket), np.max(residuals_bucket), 400)

        a, b, loc, scale = johnsonsu.fit(residuals_bucket)

        pdf = johnsonsu.pdf(x_vals, a, b, loc=loc, scale=scale)

        # plt.figure()
        # plt.hist(residuals_bucket, bins=20, density=True)
        # plt.plot(x_vals, pdf)

    else:
        a = np.nan
        b = np.nan
        loc = np.nan
        scale = np.nan

    a_vec[i] = a
    b_vec[i] = b
    loc_vec[i] = loc
    scale_vec[i] = scale

plt.figure()
plt.plot(bucket_mid, a_vec, marker='o')
plt.xlabel('Normalised Delta [-]')
plt.ylabel('Gamma Parameter [-]')

plt.figure()
plt.plot(bucket_mid, b_vec, marker='o')
plt.xlabel('Normalised Delta [-]')
plt.ylabel('Delta Parameter [-]')

plt.figure()
plt.plot(bucket_mid, loc_vec, marker='o')
plt.xlabel('Normalised Delta [-]')
plt.ylabel('Xi Parameter [$]')

plt.figure()
plt.plot(bucket_mid, scale_vec, marker='o')
plt.xlabel('Normalised Delta [-]')
plt.ylabel('Lambda Parameter [$]')

# parameter a:
model_a, residuals_a = linear_OLS(bucket_mid[~np.isnan(a_vec)], a_vec[~np.isnan(a_vec)])

# parameter b:
b_vec_clean = b_vec[~np.isnan(b_vec)]
bucket_mid_clean = bucket_mid[~np.isnan(b_vec)]

# parameter loc:
model_loc, residuals_loc = linear_OLS(bucket_mid[~np.isnan(loc_vec)], loc_vec[~np.isnan(loc_vec)])


########################################################################################################################
################################### Generate Unscaled from Scaled stochastically  ######################################
########################################################################################################################

model_base = sm.add_constant(scaled_deltas)
model_base = model_main.predict(model_base)
model_residuals = np.zeros(len(scaled_deltas))

largest = []
smallest = []

deviation = 0.05
for i in range(len(scaled_deltas)):

    val_x = np.array([[1, scaled_deltas[i]]])

    a_param = model_a.predict(val_x)[0]
    a_param = check_max_min_param(a_vec[~np.isnan(a_vec)], a_param, deviation)

    b_param = np.interp(scaled_deltas[i], bucket_mid[~np.isnan(b_vec)], b_vec[~np.isnan(b_vec)])
    b_param = check_max_min_param(b_vec[~np.isnan(b_vec)], b_param, deviation)

    loc_param = model_loc.predict(val_x)[0]
    loc_param = check_max_min_param(loc_vec[~np.isnan(loc_vec)], loc_param, deviation)

    scale_param = np.interp(scaled_deltas[i], bucket_mid[~np.isnan(scale_vec)], scale_vec[~np.isnan(scale_vec)])
    scale_param = check_max_min_param(scale_vec[~np.isnan(scale_vec)], scale_param, deviation)

    johnsonsu_dist = johnsonsu(a_param, b_param, loc=loc_param, scale=scale_param)
    model_residuals[i] = johnsonsu_dist.rvs()

Deltas_model = model_base + model_residuals

plt.figure()
plt.scatter(scaled_deltas, Deltas, marker='+', s=10, alpha=0.9, label='actuals')
plt.scatter(scaled_deltas, Deltas_model, marker='+', alpha=0.9, s=10, label='modeled')
plt.legend()
plt.xlabel('Normalised Delta [-]')
plt.ylabel('Delta [$]')

# Create the main scatter plot
fig, ax = plt.subplots()
ax.scatter(scaled_deltas, Deltas, marker='+', s=30, alpha=0.9, label='actuals')
ax.scatter(scaled_deltas, Deltas_model, marker='.', alpha=0.9, s=30, label='modeled')
ax.legend()
ax.grid()
ax.set_xlabel('Standardised Delta [$/bbl]')
ax.set_ylabel('Delta [$/bbl]')
# Define the coordinates and dimensions for the zoom-in frame
x1, x2, y1, y2 = 0.5, 2, 0, 4

# Add a rectangle to indicate the zoom area
rect = plt.Rectangle((x1, y1), x2-x1, y2-y1, edgecolor='red', facecolor='none', lw=2)
ax.add_patch(rect)

# Create the inset plot
axins = inset_axes(ax, width="45%", height="45%", loc="upper left")

# Plot the zoomed-in area in the inset plot
axins.scatter(scaled_deltas, Deltas, marker='+', s=30, alpha=0.9, label='actuals')
axins.scatter(scaled_deltas, Deltas_model, marker='.', alpha=0.9, s=30, label='modeled')
axins.set_xlim(x1, x2)
axins.set_ylim(y1, y2)

# Optionally, you can hide the x and y tick labels of the inset plot
axins.set_xticks([])
axins.set_yticks([])

# Optionally, draw lines connecting the inset plot with the zoom area
mark_inset(ax, axins, loc1=1, loc2=3, fc="none", ec="0.5")
plt.show()


########################################################################################################################
################################### Translating Model Scaled Results into Unscaled using Link ##########################
########################################################################################################################

delta_matrix_descaled = np.zeros(np.shape(delta_matrix))
dim0 = np.shape(delta_matrix)[0]
dim1 = np.shape(delta_matrix)[1]

for i in range(dim0):
    for j in range(1, dim1):
        delta_matrix_descaled[i, j] = johnsonsu_descaling(delta_matrix[i, j], model_main, model_a, model_loc, a_vec, b_vec, loc_vec, scale_vec, bucket_mid, deviation)

delta_matrix_descaled[:, 0] = delta_matrix[:, 0]

# Compute the PnL for the lossses
final_matrix_descaled = build_delta_matrix(delta_matrix_descaled)
final_matrix_descaled[:, 0] = 0

mean = np.mean(final_matrix_descaled, axis=0)
std = np.std(final_matrix_descaled, axis=0)
conf_interval = 1.96 * std / np.sqrt(final_matrix_descaled.shape[0])

plt.figure(figsize=(10, 6))
plt.plot(mean, label='Mean of de-standardised simulations')
plt.plot(mean_scaled, label='Mean of standardised simulations', linestyle='--')
plt.fill_between(np.arange(mean.shape[0]), mean - conf_interval, mean + conf_interval, color='b', alpha=0.2, label='95% Confidence Interval de-standardised')
plt.axhline(y=0.2, color='red', linestyle='dotted', linewidth=2, label=f'Trading costs without stop loss (0.20 $/bbl)')
plt.ylim([-0.4, 2.5])
plt.grid()
plt.xlabel('Days after activation')
plt.ylabel('Cumulative profit and loss of reversion position [$/bbl]')
plt.legend()
plt.show()

x_d = np.linspace(min(data)-2, max(data)+2, 10000)
log_density = kde.score_samples(x_d)
density = np.exp(log_density)

# Create a figure and two subplots
fig, (ax1, ax2) = plt.subplots(1, 2, sharey=True, figsize=(12, 6))
# Define the limits for the x-axis ranges
left_xlim = (-6, -1.5)
right_xlim = (1.5, 6)

# Plot the left portion of the data
ax1.plot(x_d, density)
ax1.hist(val_array_all, bins=int(np.sqrt(len(val_array_all))), density=True, color='gray', alpha=0.5)
ax1.axvline(-1.6, color='r', linestyle='dashed', linewidth=2, label='Threshold: -1.6')
ax1.set_xlim(left_xlim)
ax1.set_ylim([0, 0.03])
ax1.set_xlabel('Standardised Delta [-]')
ax1.set_ylabel('Density')
ax1.grid(True)
# Plot the right portion of the data
ax2.plot(x_d, density, label='Kernel Density Estimation')
ax2.hist(val_array_all, bins=int(np.sqrt(len(val_array_all))), density=True, color='gray', alpha=0.5, label='Kernel Density Sampling')
ax2.axvline(1.6, color='r', linestyle='dashed', linewidth=2, label='Threshold: 1.6')
ax2.set_xlim(right_xlim)
ax2.set_ylim([0, 0.03])
ax2.set_xlabel('Standardised Delta [-]')
ax2.grid(True)
ax2.legend()
# Adjust the layout to create a visual break
fig.subplots_adjust(wspace=0.08)
